'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Class to XXXXXXX.
-----------------------------------------------
'''
from seaborn import lineplot, histplot, heatmap, color_palette
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from typing import Dict, List, Optional, Tuple
import torch
from . interpreters import id_state

class PlotGridValues :
    
    def __init__(self, shape:tuple, dict_acciones:dict, axes_labels:Optional[List]=[[],[]]):
        assert(len(shape) == 2)
        self.shape = shape
        self.dict_acciones = dict_acciones
        self.nA = len(dict_acciones.keys())
        self.axes_labels = axes_labels
        
    def plot_policy(self, policy, V=None, ax=None):
        try:
            policy = np.flipud(np.array(policy).reshape(self.shape))
        except Exception as e:
            print('Source error:', e)
            raise Exception('Unreadable policy!')
        annotations = np.vectorize(self.dict_acciones.get)(policy)
        if V is None:
            values = np.zeros(self.shape)
        else:
            try:
                values = np.flipud(np.array(V).reshape(self.shape))
            except:
                raise Exception('Valores no legibles!')
        xticklabels = self.axes_labels[0]
        yticklabels = self.axes_labels[1]
        if ax is None:
            heatmap(
                values,
                annot=annotations,
                cbar=False,
                fmt="",
                cmap=color_palette("Blues", as_cmap=True),
                linewidths=1.7,
                linecolor="black",
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                annot_kws={"fontsize": "xx-large"},
            ).set(title="Action per state")
            plt.plot()
        else:
            heatmap(
                values,
                annot=annotations,
                cbar=False,
                fmt="",
                cmap=color_palette("Blues", as_cmap=True),
                linewidths=1.7,
                linecolor="black",
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                annot_kws={"fontsize": "xx-large"},
                ax = ax
            ).set(title="Action per state")

    def plot_V_values(self, V, ax=None):
        try:
            V = np.flipud(np.array(V).reshape(self.shape))
        except:
            raise Exception('Valores no legibles!')
        xticklabels = self.axes_labels[0]
        yticklabels = self.axes_labels[1]
        if ax is None:
            heatmap(
                V,
                annot=True,
                cbar=True,
                fmt="",
                cmap=color_palette("Blues", as_cmap=True),
                linewidths=0.7,
                linecolor="black",
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                annot_kws={"fontsize": "x-large"},
            ).set(title="V-values")
        else:
            heatmap(
                V,
                annot=True,
                cbar=True,
                fmt="",
                cmap=color_palette("Blues", as_cmap=True),
                linewidths=0.7,
                linecolor="black",
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                annot_kws={"fontsize": "x-large"},
                ax = ax
            ).set(title="V-values")

    def plot_policy_and_values(self, policy, V):
        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(12, 5))
        self.plot_policy(policy, V, ax=ax[0])
        self.plot_V_values(V, ax=ax[1])
        ax[0].set_title('Policy', fontsize='18')
        ax[1].set_title('Max values', fontsize='18')


class PlotQApprox2D :
    '''
    Plots a function approximation of Q values for a bidimensional
    state space.
    '''
    def __init__(self, 
                 Q:any, 
                 state_scales:List, 
                 dict_acciones:Dict,
                 mesh_size:Optional[int]=20,
                 use_tensors:Optional[bool]=False,
                 state_interpreter:Optional[any]=id_state) -> None:
        assert(hasattr(Q, 'predict'))
        self.Q = Q
        assert(len(state_scales) == 2)
        self.state_scales = state_scales
        self.dict_acciones = dict_acciones
        self.mesh_size = mesh_size
        self.use_tensors = use_tensors
        self.state_interpreter = state_interpreter

    def best_actions(self):
        q_table = self.create_q_table()
        # Find best actions
        policy = np.argmax(q_table, axis=2)
        # create plot object
        axes_labels = self.create_labels() 
        plot = PlotGridValues(shape=(self.mesh_size,self.mesh_size), 
                              dict_acciones=self.dict_acciones, 
                              axes_labels=axes_labels)
        plot.plot_policy(policy=policy)

    def best_values(self, max_deci:Optional[int]=2):
        q_table = self.create_q_table()
        # Find best values
        values = np.max(q_table, axis=2)
        if max_deci == 0:
            values = np.vectorize(lambda x: int(x))(values)
        else:
            values = np.vectorize(lambda x: round(x, max_deci))(values)
        # create plot object
        axes_labels = self.create_labels(max_deci=max_deci) 
        plot = PlotGridValues(shape=(self.mesh_size,self.mesh_size), 
                              dict_acciones=self.dict_acciones,
                              axes_labels=axes_labels)
        plot.plot_V_values(V=values)

    def best_actions_with_values(self, max_deci:Optional[int]=2):
        q_table = self.create_q_table()
        # Find best actions
        policy = np.argmax(q_table, axis=2)
        # Find best values
        values = np.max(q_table, axis=2)
        if max_deci == 0:
            values = np.vectorize(lambda x: int(x))(values)
        else:
            values = np.vectorize(lambda x: round(x, max_deci))(values)
        # create plot object
        axes_labels = self.create_labels(max_deci=max_deci) 
        plot = PlotGridValues(shape=(self.mesh_size,self.mesh_size), 
                              dict_acciones=self.dict_acciones,
                              axes_labels=axes_labels)
        plot.plot_policy_and_values(policy=policy, V=values)

    def create_q_table(self):
        # Create mesh from bidimensional states
        mesh_size = self.mesh_size
        X, Y = self.create_mesh()
        # Create q table from Q function
        actions = list(self.dict_acciones.keys())
        n = len(actions)
        q_table = np.zeros((mesh_size, mesh_size, n))
        for j in range(mesh_size):
            for i in range(mesh_size):
                state = (X[j], Y[i])
                if self.use_tensors:
                    state = torch.tensor(state, dtype=torch.float32)
                state = self.state_interpreter(state)
                # print(f'state:{state}')
                if hasattr(self.Q, 'values_vector'):
                    q_table[(mesh_size - 1) - i, j, :] = self.Q.values_vector(state)
                else:
                    nA = len(self.dict_acciones.keys())
                    for a in range(nA):
                        q_table[(mesh_size - 1) - i, j, a] = self.Q.predict(state, a)
        return q_table

    def create_labels(self, max_deci:Optional[int]=2):
        X, Y = self.create_mesh()
        xticklabels = [str(round(x, max_deci)) for x in X]
        yticklabels = [str(round(y, max_deci)) for y in Y]
        return [xticklabels, yticklabels]

    def create_mesh(self) -> Tuple[np.ndarray, np.ndarray]:
        x_lims = self.state_scales[0]
        X = np.linspace(x_lims[0], x_lims[1], self.mesh_size)
        y_lims = self.state_scales[1]
        Y = np.linspace(y_lims[0], y_lims[1], self.mesh_size)
        return X, Y


class PlotPolicyApprox2D :
    '''
    Plots a policy for a bidimensional state space.
    '''
    def __init__(self, policy:any, 
                 state_scales:List, 
                 dict_acciones:Dict,
                 mesh_size:Optional[int]=20) -> None:
        assert(hasattr(policy, 'predict'))
        self.policy = policy
        assert(len(state_scales) == 2)
        self.state_scales = state_scales
        self.dict_acciones = dict_acciones
        self.mesh_size = mesh_size

    def create_policy_table(self):
        # Create mesh from bidimensional states
        mesh_size = self.mesh_size
        x_lims = self.state_scales[0]
        X = np.linspace(x_lims[0], x_lims[1], mesh_size)
        y_lims = self.state_scales[1]
        Y = np.linspace(y_lims[0], y_lims[1], mesh_size)
        # Create policy table from policy
        n = len(self.dict_acciones.keys())
        policy_table = np.zeros((mesh_size, mesh_size, n))
        for j in range(mesh_size):
            for i in range(mesh_size):
                state = (X[j], Y[i])
                policy_table[(mesh_size - 1) - i, j, :] = self.policy.predict(state)
        return policy_table

    def best_actions(self):
        policy_table = self.create_policy_table()
        # Find best actions
        policy = np.argmax(policy_table, axis=2)
        # create plot object
        plot = PlotGridValues((self.mesh_size,self.mesh_size), self.dict_acciones)
        plot.plot_policy(policy=policy)

    def best_values(self, max_deci:Optional[int]=2):
        policy_table = self.create_policy_table()
        # Find best values
        values = np.max(policy_table, axis=2)
        if max_deci == 0:
            values = np.vectorize(lambda x: int(x))(values)
        else:
            values = np.vectorize(lambda x: round(x, max_deci))(values)
        # create plot object
        plot = PlotGridValues((self.mesh_size,self.mesh_size), self.dict_acciones)
        plot.plot_V_values(V=values)

    def best_actions_with_values(self, max_deci:Optional[int]=2):
        policy_table = self.create_policy_table()
        # Find best actions
        policy = np.argmax(policy_table, axis=2)
        # Find best values
        values = np.max(policy_table, axis=2)
        if max_deci == 0:
            values = np.vectorize(lambda x: int(x))(values)
        else:
            values = np.vectorize(lambda x: round(x, max_deci))(values)
        # create plot object
        plot = PlotGridValues((self.mesh_size,self.mesh_size), self.dict_acciones)
        plot.plot_policy_and_values(policy=policy, V=values)


class Plot :
    '''
    Gathers a number of frequently used visualizations.
    '''

    def __init__(self, data:pd.DataFrame):
        self.data = data

    def plot_rewards(self, file:str=None) -> plt.axis:
        '''
        Plots the reward per episode.
        Input:
            - file, string with the name of file to save the plot on.
        Output:
            - axis, a plt object, or None.
        '''
        assert('model' in self.data.columns)
        assert('environment' in self.data.columns)
        models = self.data.model.unique()
        vs_models = True if len(models) > 1 else False
        fig, ax = plt.subplots(figsize=(4,3.5))
        data = self.data.copy()
        if 'simulation' in self.data.columns:
            data = data.groupby(['model', 'environment', 'simulation', 'episode'])["reward"].sum().reset_index()
        else:
            data = data.groupby(['model', 'environment', 'episode'])["reward"].sum().reset_index()
        ax.set_xlabel('Episode')
        ax.set_ylabel('Total reward')
        ax.grid()
        if vs_models:
            ax = lineplot(x='episode', y='reward', hue='model', data=data)
        else:
            ax = lineplot(x='episode', y='reward', data=data)
        if file is not None:
            plt.savefig(file, dpi=300, bbox_inches="tight")
        return ax

    def plot_round_reward(self, file:str=None) -> plt.axis:
        '''
        Plots the reward per round.
        Input:
            - file, string with the name of file to save the plot on.
        Output:
            - axis, a plt object, or None.
        '''
        assert('model' in self.data.columns)
        assert('environment' in self.data.columns)
        models = self.data.model.unique()
        vs_models = True if len(models) > 1 else False
        fig, ax = plt.subplots(figsize=(4,3.5))
        data = self.data.copy()
        ax.set_xlabel('Round')
        ax.set_ylabel('Reward')
        ax.grid()
        if vs_models:
            ax = lineplot(x='round', y='reward', hue='model', data=data)
        else:
            ax = lineplot(x='round', y='reward', data=data)
        if file is not None:
            plt.savefig(file, dpi=300, bbox_inches="tight")
        return ax

    def plot_histogram_rewards(self, file:str=None) -> plt.axis:
        '''
        Plots a histogram with the sum of rewards per episode.
        Input:
            - file, string with the name of file to save the plot on.
        Output:
            - axis, a plt object, or None.
        '''
        assert('model' in self.data.columns)
        assert('environment' in self.data.columns)
        models = self.data.model.unique()
        vs_models = True if len(models) > 1 else False
        fig, ax = plt.subplots(figsize=(4,3.5))
        ax.set_xlabel('Sum of rewards')
        ax.set_ylabel('Frequency')
        ax.grid()
        if vs_models:
            df = self.data.groupby(['model','environment','episode']).reward.sum().reset_index()
            ax = histplot(x='reward', hue='model', data=df)
        else:
            df = self.data.groupby(['environment','episode']).reward.sum().reset_index()
            ax = histplot(x='reward', data=df)
        if file is not None:
            plt.savefig(file, dpi=300, bbox_inches="tight")
        df = self.data.groupby(['environment','model','episode']).reward.sum().reset_index()
        total_reward = df.groupby('model').reward.mean()
        print('Average sum of rewards:\n', total_reward)
        df = self.data.groupby(['environment','model','episode']).done.sum().reset_index()
        df["done"] = df["done"].astype(int)
        termination = df.groupby('model').done.mean()*100
        print('\nEpisode termination percentage:\n', termination)
        return ax


'''
© Copyright ArdillaByte Inc. 2023
-----------------------------------------------
Classes for implementing the learning methods for 
large and for continuum state spaces using Neural Networks
as approximation function for Q values.
We assume a discrete action space.
We assume epsilon-greedy action selection.
-----------------------------------------------
'''
from . BaseAgent import AgentNN
from copy import deepcopy
import torch
from torch.utils.data import Dataset, DataLoader
from typing import Dict, List, Tuple
import numpy as np
from pathlib import Path


class ExperienceDataset(Dataset):
    '''s
    Creates the dataset out of the experience stream
    '''
    def __init__(
            	self, 
                states:List[torch.Tensor], 
                actions:List[int], 
                updates:List[float]
            ) -> None:
        self.states = states
        self.actions = actions
        self.updates = [torch.tensor(u, dtype=torch.float32) for u in updates]
        n = len(self.states)
        assert (len(self.actions) == n)
        assert (len(self.updates) == n)

    def __len__(self):
        return len(self.states)

    def __getitem__(self, idx:int):
        # print(type(self.states[idx]))
        # print(type(self.actions[idx]))
        # print(type(self.updates[idx]))
        state = self.states[idx].to(torch.float32)
        action = int(self.actions[idx])
        update = self.updates[idx].to(torch.float32) 
        return state, action, update



class DQN(AgentNN) :
    '''
    Implements the Deep Q Network with 
    experience replay and target network.
    '''
    def __init__(self, parameters:Dict[str, any]) -> None:
        super().__init__(parameters)
        self.target_network_latency = parameters["target_network_latency"]
        self.len_exp = parameters["len_exp"]
        self.batch_size = parameters["batch_size"]
        self.num_epochs = parameters["num_epochs"]
        # Create target network
        self.NN_hat = deepcopy(self.NN)
        self.learn_debug = False

    def update(
            	self, 
                next_state:torch.Tensor, 
                reward:float, 
                done:bool
            ) -> None:
        '''
        Agent updates NN with experience replay and updates target NN.
        '''
        n = len(self.actions)
        k = self.len_exp
        # Obtain length of experience
        if n < k:
            # print('agent only learns with enough experience')
            #agent only learns with enough experience
            pass
        else:
            # Create the experience stream with random indices from the whole history
            states, actions, next_states, rewards, dones = self.create_experience_stream(next_state, reward, done)
            # Create the dataset and dataloader
            ds = self.create_dataset(states, actions, next_states, rewards, dones)
            # Train for number of epochs
            # print('Training...')
            for e in range(self.num_epochs):
                ds_loader = DataLoader(ds, batch_size=self.batch_size, shuffle=True)
                self.NN.learn(ds_loader)
            # Check if it's turn to update the target network
            # if len(self.actions) % self.target_network_latency == 0 or done:
            if len(self.actions) % self.target_network_latency == 0:
                # print('Updating Target Network')
                self.NN_hat = deepcopy(self.NN)

    def create_experience_stream(
            	self, 
                next_state:any, 
                reward:float, 
                done:bool
            ) -> Tuple[
					List[any], 
					List[int], 
					List[any], 
					List[float], 
					List[bool]
	            ]:
        # Get records
        states = list(self.states)[:-1].copy()
        actions = list(self.actions)[:-1].copy()
        next_states = list(self.states)[1:].copy()
        rewards = list(self.rewards)[1:].copy()
        dones = list(self.dones)[1:].copy()
        # Check lists are in good shape
        lengths = f'#states:{len(self.states)} -- #actions:{len(self.actions)} -- #rewards:{len(self.rewards)} -- #dones:{len(self.dones)}'
        n = len(self.states)
        assert(n > 1), 'Error: len_experience should be at least 2'
        assert(len(self.actions) == n), lengths
        assert(len(self.rewards) == n), lengths
        assert(len(self.dones) == n), lengths
        # Create mask with random indices
        mask = np.random.choice(range(len(self.states) - 1), 
                                size=self.len_exp - 1, 
                                replace=False)
        tensor_mask = torch.tensor(mask).to(torch.int32)
        # print('mask:', tensor_mask, type(tensor_mask))
        # Get the randomly selected states
        stacked_states = torch.stack([state for state in self.states])
        filtered_states = torch.index_select(stacked_states, 0, tensor_mask)
        list_of_states = list(torch.unbind(filtered_states, dim=0))
        states = list_of_states + [self.states[-1]]
        # Get the randomly selected NEXT states
        mask_next = (np.array(mask) + 1).tolist()
        tensor_mask = torch.tensor(mask_next)
        next_states = list_of_states + [next_state]
        # Get the randomly selected actions
        actions = np.array(self.actions.copy())[mask].tolist() + [self.actions[-1]]
        # Get the randomly selected rewards
        rewards = np.array(self.rewards.copy())[mask_next].tolist() + [reward]
        # Get the randomly selected dones
        dones = np.array(self.dones.copy())[mask_next].tolist() + [done]
        # Make sure indices don't correspond to terminal satates
        states, actions, next_states, rewards, dones = self.avoid_indices_for_terminal_states(states, actions, next_states, rewards, dones)
        # Check lists are in good shape
        self._check_lists_in_good_shape(states, actions, next_states, rewards, dones)
        return states, actions, next_states, rewards, dones

    def avoid_indices_for_terminal_states(
            	self,
                states:List[torch.Tensor], 
                actions:List[int], 
                next_states:List[torch.Tensor], 
                rewards:List[float], 
                dones:List[bool]
            ) -> Tuple[
                	List[any], 
                    List[int], 
                    List[any],
                    List[float], 
                    List[bool]
                ]:
        indices_no_terminal_states = [i for i, a in enumerate(actions) if not np.isnan(a)]
        # Filtering lists
        actions = np.array(actions.copy())[indices_no_terminal_states].tolist()
        rewards = np.array(rewards.copy())[indices_no_terminal_states].tolist()
        dones = np.array(dones.copy())[indices_no_terminal_states].tolist()
        # Filtering lists of tensors
        tensor_mask = torch.tensor(indices_no_terminal_states)
        #     list of states
        stacked_states = torch.stack(states)
        filtered_states = torch.index_select(stacked_states, 0, tensor_mask)
        states = list(torch.unbind(filtered_states, dim=0))
        #     list of next_states
        stacked_states = torch.stack(next_states)
        filtered_states = torch.index_select(stacked_states, 0, tensor_mask)
        next_states = list(torch.unbind(filtered_states, dim=0))
        return states, actions, next_states, rewards, dones

    def create_dataset(self, 
                states:List[torch.Tensor], 
                actions:List[int], 
                next_states:List[torch.Tensor], 
                rewards:List[float], 
                dones:List[bool]
            ) -> ExperienceDataset:
        updates = [self.get_update(next_states[i], rewards[i], dones[i]) for i in range(len(rewards))]
        states, actions, updates = self.filter_ds(states, actions, updates)
        n = len(states)
        assert(n > 0)
        return ExperienceDataset(states, actions, updates)
               
    def filter_ds(self, states, actions, updates):
        mask = [i for i, state in enumerate(states) if state[0] != 0]
        tensor_mask = torch.tensor(mask).to(torch.int32)
        stacked_states = torch.stack([state for state in states])
        filtered_states = torch.index_select(stacked_states, 0, tensor_mask)
        list_of_states = list(torch.unbind(filtered_states, dim=0))
        return list_of_states, np.array(actions)[mask].tolist(), np.array(updates)[mask].tolist()

    def get_update(
                self, 
                next_state:torch.Tensor, 
                reward:float, 
                done:bool
            ) -> float:
        if done:
            # Episode is finished. No need to bootstrap update
            G = reward
            if self.learn_debug:
                print(f'G:{G} --- reward:{reward} --- ENV DONE')
        else:
            # Episode is active. Bootstrap update using Target Network
            Qvals = self.NN_hat.values_vector(next_state)
            maxQ = max(Qvals)
            G = reward + self.gamma * maxQ
            if self.learn_debug:
                print(f'G:{G} --- reward:{reward} --- gamma:{self.gamma} --- maxQ:{maxQ}')
        return G    
    
    def restart(self) -> None:
        # Override restart to preserve experience
        if len(self.rewards) > 1:
            # print('Including empty rewards and dones at begining of episode')
            self.rewards.append(np.nan)
            self.dones.append(np.nan)
        self.NN.restart()

    def get_records(self) -> Tuple[
									List[any],
									List[int],
									List[any],
									List[float],
									List[bool]
                                ]:
        # Get records
        states = list(self.states)[:-1].copy()
        actions = list(self.actions)[:-1].copy()
        next_states = list(self.states)[1:].copy()
        rewards = list(self.rewards)[1:].copy()
        dones = list(self.dones)[1:].copy()
        self._check_lists_in_good_shape(states, actions, next_states, rewards, dones)
        # Make sure indices don't correspond to terminal satates
        states, actions, next_states, rewards, dones = self.avoid_indices_for_terminal_states(states, actions, next_states, rewards, dones)
        self._check_lists_in_good_shape(states, actions, next_states, rewards, dones)
        return states, actions, next_states, rewards, dones

    def print_dataset(self) -> None:
        # Download records
        states, actions, next_states, rewards, dones = self.get_records()
        # Create the dataset
        ds = self.create_dataset(states, actions, next_states, rewards, dones)
        # Create the dataloader
        ds_loader = DataLoader(ds, batch_size=1, shuffle=False)
        for state, action, update in ds_loader:
            print('')
            print('='*30)
            print('state:')
            print(state)
            print('-'*20)
            print(f'Action:\n\t{action}')
            print('-'*20)
            print(f'Update:\n\t{update}')

    def _check_lists_in_good_shape(
                self,
                states:List[any],
                actions:List[int],
                next_states:List[any],
                rewards:List[float],
                dones:List[bool]
            ) -> None:
        # Check lists are in good shape
        lengths = f'#states:{len(states)} -- #actions:{len(actions)} #next_states:{len(next_states)} -- #rewards:{len(rewards)} -- #dones:{len(dones)}'
        n = len(states)
        assert(len(next_states) == n), lengths
        assert(len(actions) == n), lengths
        assert(len(rewards) == n), lengths
        assert(len(dones) == n), lengths






'''
© Copyright ArdillaByte Inc. 2023
-----------------------------------------------
Classes for implementing the Q functions with neural networks
-----------------------------------------------
'''
import torch
from torch.utils.data import DataLoader
import numpy as np
from pathlib import Path
from . networks import MLP
from copy import deepcopy
from prettytable import PrettyTable

class NNasQ :
    def __init__(self, parameters):
        self.parameters = parameters
        self.NN = parameters["NN"]
        network_parameters = self.NN.model.parameters()
        self.nA = parameters["nA"]
        self.loss_fn = parameters["loss_fn"]
        alpha = parameters["alpha"]
        optimizer_class = parameters["optimizer_class"]
        if optimizer_class == 'Adam':
            self.optimizer = torch.optim.Adam(network_parameters, lr=alpha)
        else:
            raise Exception(f'Optimizer class {optimizer_class} not implemented!')
        self.backup_NN = deepcopy(self.NN)
        self.losses = []

    def predict(self, state, action):
        with torch.no_grad():
            # Get predicted Q values
            Qs = self.NN(state)
            if len(Qs.shape) > 1:
                Qs.squeeze()
            # Transform to list
            Qs = Qs.data.tolist()
        return Qs[action]
    
    def values_vector(self, state):
        with torch.no_grad():
            # Get predicted Q values
            Qs = self.NN(state)
            if len(Qs.shape) > 1:
                Qs.squeeze()
            # Transform to list
            Qs = Qs.data.tolist()
        return Qs
    
    def learn(self, ds_loader:DataLoader):
        '''
        Trains the NN with the given dataset
        '''
        for batch_states, batch_actions, batch_updates in ds_loader:
            # print('')
            # print(f'batch_states:{batch_states}')
            # print(f'batch_actions:{batch_actions}')
            # print(f'batch_updates:{batch_updates}')
            # Clear the gradient
            self.optimizer.zero_grad()
            # Get the batch predicted values
            batch_qvals = self.NN.forward(batch_states)
            # print(f'batch_qvals:{batch_qvals}')
            # Confirm the number of actions
            nA = self.parameters["nA"]
            # Get the indices for the actions
            mask = torch.tensor([i * nA + a for i, a in enumerate(batch_actions)], dtype=torch.long)
            # Keep only the q values corresponding to the actions
            batch_X = torch.take(batch_qvals, mask)
            # Adapt the update
            batch_Y = batch_updates
            # Determine loss
            # print(f'X:{batch_X} --- Y:{batch_Y}')
            loss = self.loss_fn(batch_X, batch_Y)
            self.losses.append(loss.item())
            # print('loss:', loss.item())
            # Find the gradients by backward propagation
            loss.backward()
            # Update the weights with the optimizer
            self.optimizer.step()

    def save(self, file:Path):
        torch.save(self.NN.state_dict(), file)

    def load(self, file:Path):
        self.NN.load_state_dict(torch.load(file))

    def restart(self):
        pass

    def reset(self):
        self.restart()
        # Instantiate original model
        self.NN = deepcopy(self.backup_NN)
        network_parameters = self.NN.model.parameters()
        # Create optimizer
        alpha = self.parameters["alpha"]
        optimizer_class = self.parameters["optimizer_class"]
        if optimizer_class == 'Adam':
            self.optimizer = torch.optim.Adam(network_parameters, lr=alpha)

    def summary(self):
        table = PrettyTable(['Modules', 'Parameters'])
        total_params = 0
        for name, parameter in self.NN.model.named_parameters():
            if not parameter.requires_grad: continue
            params = parameter.numel()
            table.add_row([name, params])
            total_params+=params
        print(table)
        print(f'Total Trainable Params: {total_params}')


class MLPasQ(NNasQ):
    '''
    Defines a Q function using a Multi-layer Perceptron
    '''
    def __init__(self, parameters):
        self.sizes = parameters["sizes"]
        intermediate_activation_function = parameters["intermediate_activation_function"]
        last_activation_function = parameters["last_activation_function"]
        # Create Multi-layer Perceptron
        NN = MLP(sizes=self.sizes,
                 intermediate_activation_function=intermediate_activation_function, 
                 last_activation_function=last_activation_function)
        parameters["NN"] = NN
        # Define loss function as Mean Square Error
        parameters["loss_fn"] = torch.nn.MSELoss()
        super().__init__(parameters)
       



'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Clase para correr el campo de pozos con sus
clusters y camiones.
-----------------------------------------------
'''
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import networkx as nx
import pandas as pd
import numpy as np
from typing import List, Dict, Optional, Tuple, Union
from IPython.display import HTML
from copy import deepcopy
import torch
from prettytable import PrettyTable
import pickle
from pathlib import Path

from chaparral.modelos.clases import (
	Cluster, 
	Camion, 
	Ruta, 
	PseudoRutaCamionEstacionario
)
from chaparral.modelos.constantes import (
	FACTOR_RECOMPENSA,
	RUIDO,
	VALOR_CRITICO,
	FACTOR_RENDER
)


class CampoPozos:
	'''
	El entorno de los clusters unidos por caminos.
 
	Atributos:
		- Lista de clusters
		- Lista de caminos
		- Lista de camiones
		- Datos de las rutas
	'''
	def __init__(
				self, 
				clusters:List[Cluster], 
				datos_rutas:Dict[str,float],
				camiones:List[Camion]
			) -> None:
		# -------------------------------
		# Datos provenientes de clusters
		# -------------------------------
		self.clusters = clusters
		self.backup_clusters = deepcopy(clusters)
		self.num_clusters = len(self.clusters)
		self.punto_partida = clusters[0].nombre
		self.dict_clusters = {cluster.nombre:p for p, cluster in enumerate(self.clusters)}
		self.coordenadas = {cluster.nombre:cluster.coordenadas for cluster in self.clusters}
		# -------------------------------
		# Datos provenientes de rutas
		# -------------------------------
		self.datos_rutas = datos_rutas
		# -------------------------------
		# Datos provenientes de camiones
		# -------------------------------
		self.camiones = camiones
		for camion in self.camiones:
			camion.ubicacion = 0
			camion.ruta = None
		self.backup_camiones = deepcopy(camiones)
		self.num_camiones = len(self.camiones)
		# -------------------------------
		# Inicializa el estado
		# -------------------------------
		self.estado = self.reset()
		self.nS = self.estado.shape[0]
		self.comienzo = True # El entorno recién se inicializa
		# -------------------------------
		# Crea el bookkeeping de las acciones
		# -------------------------------
		# Las acciones posibles es decirle a cada camión
		# a qué cluster ir
		self.nA = self.num_clusters * self.num_camiones
		# -------------------------------
		# Para usar en debugging
		# -------------------------------
		self.debug = False
		# -------------------------------
		# Para usar en render
		# -------------------------------
		self.render_mode = 'rgb_array'

	def reset(self):
		'''
		Reinicia el entorno y devuelve el estado inicial
		'''
		self.clusters = deepcopy(self.backup_clusters)
		self.camiones = deepcopy(self.backup_camiones)
		estado = self.obtener_estado()
		self.comienzo = True # El entorno recién se inicializa
		return estado
	
	def step(self, accion:int) -> Tuple[torch.tensor, float, bool]:
		'''
		Avanza el entorno de acuerdo a la tupla de acciones.

		Input:
			- accion, índice de la lista de qué ruta toma cada camión

		Output:
			- tripla con nuevo estado, recompensa y done
		'''
		# El entorno comenzó
		self.comienzo = False
		# Decodifica la acción como índice de lista de movimientos
		movimientos_camiones = self.accion_a_movimientos(accion)
		# Verifica que todos los movimientos sean válidos
		verificacion_movimientos = self._check_movimientos(movimientos_camiones)
		if not verificacion_movimientos:
			# Alguno de los movimientos no es permitido.
			# El entorno debe seguir en el mismo estado 
			# y devolver castigo mediante la tupla: 
			# (mismo estado, recompensa negativa, entorno sigue activo)
			return self.estado, -1000, False
		# Si un camión debe comenzar una nueva ruta, inicializar en la lista atributo
		self._inicializar_rutas(movimientos_camiones)
		# Obtenemos cuántas unidades temporales se demora cada
		# camión en su respectiva ruta
		unidades = self._obtener_unidades_temporales()
		# Las unidades temporales a transcurrir son
		# el menor tiempo para terminar una ruta
		if self.debug:
			print('Unidades temporales de las rutas:', unidades)
		unidades_temporales = min(unidades)
		# para bookkeeping de espacio de estados
		self._unidades_temporales = unidades_temporales
		if unidades_temporales == np.inf:
			# Si el mínimo es inf, significa que ningún camión se mueve
			# Esto solo debe ocurrir cuando el entorno termina
			return self.estado, -10000, True
		if self.debug:
			print(f'Unidades temporales que transcurren: {unidades_temporales}')
		# Actualiza el estado
		self._actualizar_estado(unidades_temporales)
		nuevo_estado = self.obtener_estado()
		# Encuentra la recompensa y el done
		recompensa, done = self._encontrar_recompensa_y_done(unidades_temporales)
		return nuevo_estado, recompensa, done
	
	def obtener_estado(
				self,
				noise_:Optional[float]=RUIDO
			) -> torch.tensor:
		'''
		Obtiene el estado actual a partir de los atributos del entorno
		'''
		tensores_clusters = [cluster.cluster_a_tensor() for cluster in self.clusters[1:]]
		tensores_camiones = [camion.camion_a_tensor(self.num_clusters) for camion in self.camiones]
		tensores = tensores_clusters + tensores_camiones
		tensores = torch.cat(tensores)
		# Incluimos algo de ruido en el tensor
		shape = tensores.shape
		RUIDO = torch.rand(shape) * noise_
		return torch.add(tensores, RUIDO)

	def _actualizar_estado(
				self, 
				unidades_temporales:float
			) -> None:
		'''
		Actualiza el estado con la lógica siguente:
			1. Actualiza los clusters
			2. Actualiza los camiones en su respectiva ruta
			3. Si un camión llega al final de la ruta, descarga lo que pueda
		'''
		# 1. Pedimos a cada cluster que actualize con la cantidad de unidades temporales
		for cluster in self.clusters:
			cluster.actualizar(unidades_temporales)
		# 2. Pedimos a cada ruta que actualize con la cantidad de unidades temporales
		for camion in self.camiones:
			camion.ruta.actualizar(unidades_temporales)
			# 3. Verificar si el camión que va en la ruta llegó al final
			if camion.ruta.fin_recorrido():
				nombre_cluster_llegada = camion.ruta.objetivo.nombre
				if self.debug:
					print(f'Camión {camion.nombre} llegó a cluster {nombre_cluster_llegada}')
				# Descarga químicos en cluster
				indice_cluster = self.dict_clusters[nombre_cluster_llegada]
				self._descargar_camion_en_cluster(
					nombre_cluster=nombre_cluster_llegada, 
					camion=camion
				)
				# Actualizar la posición del camión
				camion.ubicacion = self.dict_clusters[nombre_cluster_llegada]
				# Dejar en espera de nueva ruta para el camión
				camion.ruta = None

	def _encontrar_recompensa_y_done(self, unidades_temporales) -> Tuple[float, bool]:
		'''Encuentra la recompensa en el estado obtenido'''
		# Chequeamos si la cantidad de algún químico en algún cluster es negativa
		for cluster in self.clusters[1:]:
			for pozo in cluster.pozos:
				for quimico in pozo.quimicos:
					if quimico.nivel_actual < quimico.velocidad_consumo * VALOR_CRITICO:
						if self.debug:
							print('¡Se infringieron los porcentajes críticos!')
							for cluster in self.clusters:
								print(f'\tCluster {cluster.nombre}--{pozo.nombre}--{quimico.nombre} con niveles: {quimico.nivel_actual}')
						# Por lo menos un químico está por debajo del nivel crítico => penalizar al agente
						return -1000, True
		# Chequeamos si todos los camiones están en bodega
		# y no es el estado incial
		if not self.comienzo and np.all([camion.ubicacion == 0 for camion in self.camiones]):
			# Entorno terminado
			# # Encontramos la recompensa como la suma de las diferencias 
			# # entre los niveles actuales y los requeridos (todos estos valores
			# # deben ser negativos)
			# recompensas = [np.subtract(cluster.niveles_actuales, 1) for cluster in self.clusters if cluster.activo]
			# Encontramos la recompensa como la suma de los niveles por encima del 1%
			recompensas = [quimico.nivel_actual // quimico.velocidad_consumo for cluster in self.clusters for pozo in cluster.pozos for quimico in pozo.quimicos]
			if self.debug:
				print(f'Las diferencias de químicos son: {recompensas}')
			recompensa = sum(recompensas) * FACTOR_RECOMPENSA
			return recompensa, True
		return -1 * unidades_temporales, False	

	def _obtener_unidades_temporales(self) -> Union[None, List[int]]:
		'''
		Obtiene la cantidad de unidades temporales requeridas 
		para terminar cada ruta de acuerdo a los movimientos
		de los camiones dado en el argumento

		Input:
			- movimientos_camiones, la lista de qué ruta toma cada camión

		Output:
			- Lista de unidades temporales o None en caso de acción inválida
		'''
		unidades = list()
		for camion in self.camiones:
			unidad = camion.ruta.obtener_unidades_temporales_faltantes()
			unidades.append(unidad)
		return unidades

	def _inicializar_rutas(self, movimientos_camiones:List[int]) -> None:
		'''Verifica si camión está en la ruta o si no inicializa'''
		for indice_camion, movimiento in enumerate(movimientos_camiones):
			ruta = self.camiones[indice_camion].ruta
			if ruta is None:
				self._iniciar_ruta(indice_camion, movimiento)

	def _iniciar_ruta(self, indice_camion, movimiento) -> None:
		'''Obtiene la ruta que debe seguir el camión y modifica la lista atributo'''
		# Obtiene dónde está el camión
		indice_inicial = self.camiones[indice_camion].ubicacion
		inicial = self.clusters[indice_inicial]
		# Obtiene para dónde va el camión de acuerdo al movimiento
		objetivo = self.clusters[movimiento]
		# Verifica si el movimiento estipula que el camión debe quedarse donde está
		if inicial.nombre != objetivo.nombre:
			# Crea el nombre de la ruta
			nombre = f'{inicial.nombre} <-> {objetivo.nombre}'
			# Encuentra longitud y velocidad promedio
			longitud = self.datos_rutas[nombre]['longitud']
			velocidad_promedio = self.datos_rutas[nombre]['velocidad_promedio']
			# Crea la ruta
			ruta = Ruta(
				inicial=inicial,
				objetivo=objetivo,
				longitud=longitud,
				velocidad_promedio=velocidad_promedio
			)
		else:
			# Debemos crear una pseudoruta porque el movimiento
			# determina que el camión se queda estacionario
			ruta = PseudoRutaCamionEstacionario(
				inicial=inicial,
				objetivo=objetivo,
				longitud=np.Inf,
				velocidad_promedio=0
			)
		self.camiones[indice_camion].ruta = ruta

	def _check_movimientos(self, movimientos_camiones:List[int]) -> bool:
		'''
		Verifica que todos los movimientos sean válidos
		'''
		for indice_camion, movimiento in enumerate(movimientos_camiones):
			verificacion = self._check_accion(
				indice_camion=indice_camion,
				movimiento=movimiento,
				indice_cluster=self.camiones[indice_camion].ubicacion
			)
			if not verificacion:
				return False
		return True

	def _check_accion(self, indice_camion:int, movimiento:int, indice_cluster:str) -> bool:
		'''
		Verifica que el movimiento es posible en el cluster dado
		y, si el camión está en ruta, verifica que el movimiento
		coincida con la dirección de la ruta. También es posible
		que el camión se quede estacionario
		'''
		nombre_cluster = self.clusters[indice_cluster]
		# Obtener lista de clusters accesibles desde donde está el camión
		clusters_posibles = self.clusters[indice_cluster].vecinos + [nombre_cluster]
		# Obtiene el nombre del culster dado por el movimiento
		a_cluster = self.clusters[movimiento].nombre
		# El movimiento debe estar en rango de movimientos posibles
		primer_check = a_cluster in clusters_posibles
		# Si camión en transito, accion debe coincidir con destino
		ruta = self.camiones[indice_camion].ruta
		if ruta is None:
			# Camión no está en tránsito
			segundo_check = True
		else:
			# Camión en tránsito
			destino = ruta.objetivo.nombre
			segundo_check = destino == a_cluster
		if self.debug:
			print(f'Movimiento permitido: {primer_check}, Movimiento en ruta: {segundo_check}')
		return np.all([primer_check, segundo_check])
	
	def _descargar_camion_en_cluster(
				self, 
				nombre_cluster:str,
				camion:Camion
			) -> None:
		'''Descarga los químicos en el cluster'''
		# Carga el objeto cluster
		indice_cluster = self.dict_clusters[nombre_cluster]
		cluster = self.clusters[indice_cluster]
		# Itera sobre los pozos del cluster
		for pozo in cluster.pozos:
			# Itera sobre los químicos del pozo
			for quimico in pozo.quimicos:
				# Determina si el camion tiene el químico
				indice_quimico = quimico.in_carga(camion.carga_quimicos)
				if indice_quimico is not None:
					# Encuentra la cantidad que falta del químico
					# como la diferencia entre el valor requerido y el actual
					falta = quimico.velocidad_consumo * VALOR_CRITICO - quimico.nivel_actual
					carga = camion.carga_quimicos[indice_quimico].nivel_actual
					# Calcula la cantidad a descargar del camión
					if 0 <= falta <= carga:
						descarga = falta
					elif falta > carga:
						descarga = carga
					else:
						raise Exception(f'¡Valor de descarga inválido!\n\t carga:{carga} --- falta:{falta}')
					if self.debug:
						print(f'El camión {camion.nombre} descarga {descarga} (faltaba {falta})')
					# Actualiza los niveles de cluster y camión
					quimico.nivel_actual += descarga
					camion.carga_quimicos[indice_quimico].nivel_actual -= descarga
	
	def accion_a_movimientos(self, accion:int) -> List[int]:
		"""Toma una acción y devuelve la lista correspondiente"""
		shape = tuple([self.num_clusters] * self.num_camiones)
		return list(np.unravel_index(accion, shape))
	
	def movimientos_a_accion(self, movimientos:Tuple[int]) -> int:
		"""Toma una lista de movimientos y devuelve la acción como índice"""
		shape = tuple([self.num_clusters] * self.num_camiones)
		return np.ravel_multi_index(movimientos, shape)
		
	def _list_edges(self) -> List[Tuple[str,str,Dict]]:
		if hasattr(self, 'list_edges'):
			return self.list_edges
		else:
			list_edges = [self._obtener_tupla_distancia(c1.nombre, c2) for c1 in self.clusters for c2 in c1.vecinos]
			self.list_edges = list_edges
			return self.list_edges

	def _obtener_tupla_distancia(self, cl1:str, cl2:str) -> float:
		distancia = f"{self.datos_rutas[f'{cl1} <-> {cl2}']['longitud']}km"
		velocidad = f"{self.datos_rutas[f'{cl1} <-> {cl2}']['velocidad_promedio']}km/h"
		rotulo = f'{distancia} a {velocidad}'
		dict_distancia = {'distancia':rotulo}
		return (cl1, cl2, dict_distancia)

	def generar_rutas_de_politica(
				self, 
				agente:any, 
				max_steps:Optional[int]=100
			) -> List[Tuple[Dict,PrettyTable]]:
		"""
		Genera las rutas usando un agente
		
		Input:
			- politica, una función que genera movimientos de los camiones en un estado dado
		
		Output:
			- dict_rutas, un diccionario con la ruta de cada camión
		"""
		# Hacemos un backup del valor de epsilon
		backup_epsilon = deepcopy(agente.epsilon)
		# Apagamos la exploración del agente
		agente.epsilon = 0
		# Iniciamos los registros
		num_steps = 0
		done = False
		estado = self.reset()
		accion = agente.make_decision(estado)
		movimientos_camiones = self.accion_a_movimientos(accion)
		dict_rutas = {camion.nombre:[movimientos_camiones[indice_camion]] for indice_camion, camion in enumerate(self.camiones)}
		while not done:
			num_steps += 1
			estado, recompensa, done = self.step(accion)
			accion = agente.make_decision(estado)
			movimientos_camiones = self.accion_a_movimientos(accion)
			for indice_camion, camion in enumerate(self.camiones):
				dict_rutas[camion.nombre].append(movimientos_camiones[indice_camion])
		table = PrettyTable(['Camión', 'Cluster'])
		for indice_camion in range(self.num_camiones):
			camion = self.camiones[indice_camion].nombre
			a_cluster = self.clusters[0].nombre
			table.add_row([camion, a_cluster])
		for indice_camion in range(self.num_camiones):
			for i in range(num_steps):
				camion = self.camiones[indice_camion].nombre
				indice_cluster = dict_rutas[camion][i]
				a_cluster = self.clusters[indice_cluster].nombre
				table.add_row([camion, a_cluster])
		# Reinstauramos el valor de epsilon
		agente.epsilon = backup_epsilon
		return (dict_rutas, table)
	
	def render(self):
		G = nx.Graph()
		G.add_nodes_from(self.coordenadas.keys())
		list_edges = self._list_edges()
		print(list_edges)
		G.add_edges_from(list_edges, color='red')
		order = {cluster:self._transformar_coordenadas(self.coordenadas[cluster]) for cluster in self.coordenadas.keys()}
		options = {
			'node_color': 'green',
			'node_size': 10,
			'verticalalignment':'bottom',
			'edge_color': 'red',
			'width': 0.5,
			'font_size': 8
		}
		fig = plt.figure(figsize=(15,15))
		ax = fig.add_subplot()
		nx.draw_networkx(G, pos=order, font_weight='bold', **options)
		# edge_labels = nx.get_edge_attributes(G, 'distancia')
		# nx.draw_networkx_edge_labels(G, order, edge_labels)
		# for c, camion in enumerate(self.camiones):
		# 	camion_en_pozo = camion.ubicacion
		# 	coordenadas = self.clusters[camion_en_pozo].coordenadas
		# 	circle1 = patches.Circle(coordenadas, radius=5, color='green')
		# 	ax.add_patch(circle1)
		ax.axis('off')
		fig.canvas.draw()
		string = fig.canvas.renderer.buffer_rgba()
		array = np.array(string)
		if self.render_mode == 'rgb_array':
			return array
		else:
			plt.show()

	def _transformar_coordenadas(
				self,
				coordenadas:List[float]
			) -> List[float]:
		return [coordenadas[0] * FACTOR_RENDER, coordenadas[1] * FACTOR_RENDER]

	def __str__(self) -> str:
		cadena = '<table> <tbody>'
		for cluster in self.clusters:
			cadena += str(cluster)
		cadena += '</tbody> </table>'
		cadena += '<tr> <td>&nbsp;</td> <td>&nbsp;</td> <td>&nbsp;</td> </tr>'
		cadena += '<table> <tbody>'
		for camion in self.camiones:
			cadena += str(camion)
		cadena += '</tbody> </table>'
		return cadena

	def state_visualizer(self, state):
		"""Retorna una representación visual del estado"""
		html_table = str(self)
		table_MN = pd.read_html(html_table)
		return table_MN

	def action_visualizer(self, action):
		"""Retorna una representación visual de la acción"""
		movimientos_camiones = self.accion_a_movimientos(action)
		lista_movimientos = [self.clusters[indice_cluster].nombre for indice_cluster in movimientos_camiones]
		return lista_movimientos
	
	def save(
				self,
				archivo:Optional[Union[None, str]]=None
			) -> None:
		directorio = Path.cwd() / Path('..').resolve() / Path('data', 'entornos')
		directorio.mkdir(parents=False, exist_ok=True)
		if archivo is not None:
			path_archivo = directorio / Path(archivo)
		else:
			path_archivo = directorio / Path('entorno.pkl')
		with open(path_archivo, 'wb') as file:
			pickle.dump(self, file)
			file.close()


import re
import jsonlines

import polars as pl

from chaparral.mapa.convenciones import CLUSTER2MAPA


POZOS = []
for pozo in CLUSTER2MAPA.values():
    POZOS.extend(pozo.split('/'))


def estandarizar_nombre(nombre):
    nombre = nombre.\
        replace('CHICHIMENE', 'CH').\
        replace(':1', '').\
        replace('CLUSTER', 'CL')
    cluster = re.search(r'\w+-\d+(\w+)', nombre)
    if cluster:
        return cluster.group().strip()


def load_inventario(data: pl.DataFrame) -> pl.DataFrame:
    puntos = []
    for row in data.iter_rows(named=True):
        pozo = estandarizar_nombre(row['Punto'])
        if pozo:
            quimicos = row['Producto'].split()
            for quimico in quimicos:
                puntos.append({
                    'punto': pozo,
                    'quimico': quimico,
                    'nivel_actual': row['Saldo Gln'],
                    'velocidad_consumo': row['Gln/día'],
                    'dias_con_stock': row['Stock en días'],
                })
        else:
            print(f'no se halla punto {pozo}')
    return pl.DataFrame(puntos)


if __name__ == "__main__":
    file_path = '../../data/Inventario_Pozos.csv'
    status_pozos_path = '../../data/status_pozos.csv'
    out_path = '../../data/objetos_mapa/inventario.ndjson'
    df = pl.read_csv(file_path)
    status_pozos = pl.read_csv(status_pozos_path, infer_schema_length=10000)
    status_pozos = status_pozos.with_columns(pl.col('POZO').map_elements(lambda x: estandarizar_nombre(x)))
    inventario = load_inventario(df)
    print(inventario.shape)
    inventario = inventario.join(status_pozos, left_on='punto', right_on='POZO', how='outer').\
        filter(pl.col('Estado') == 'ON').drop(['CL', 'Estado'])
    print(inventario.shape)
    inventario.write_ndjson(out_path)




'''
© Copyright ArdillaByte Inc. 2023
-----------------------------------------------
Helper functions to gather, process and visualize data

Includes:
    - Episode, Runs the environment for a number of rounds and keeps tally of everything.
    - Experiment, Compares given models on a number of measures.
    - EnvfromGameAndPl2
-----------------------------------------------
'''

import numpy as np
import pandas as pd
from tqdm.auto import tqdm
from copy import deepcopy
from chaparral.utils.plot_utils import Plot
import matplotlib.pyplot as plt
from IPython.display import clear_output, display
from time import sleep
from typing import Union, List, Tuple
from chaparral.utils.interpreters import id_state
from prettytable import PrettyTable
from gymnasium.utils.save_video import save_video
from collections import deque


class Episode :
    '''
    Runs the environment for a number of rounds and keeps tally of everything.
    '''

    def __init__(
                self, 
                environment:any,
                env_name:str, 
                agent:any, 
                model_name:str, 
                num_rounds:int, 
                id:int=0,
                state_interpreter=id_state,
            ) -> None:
        self.environment = environment
        self.env_name = env_name
        self.agent = agent
        self.model_name = model_name
        self.num_rounds = num_rounds
        self.done = False
        self.T = 1
        self.id = id
        self.sleep_time = 0.3
        self._clean_state = state_interpreter
        state_ = self.environment.reset()
        state = self._clean_state(state_)
        self.initial_state = state
        if agent is not None:
            self.agent.restart()
            self.agent.states.append(state)
            # lengths = f'#states:{len(self.agent.states)} -- #actions:{len(self.agent.actions)} -- #rewards:{len(self.agent.rewards)} -- #dones:{len(self.agent.dones)}'
            # print('-->', lengths)

    def play_round(self, verbose:int=0, learn:bool=True) -> None:
        '''
        Plays one round of the game.
        Input:
            - verbose, to print information.
                0: no information
                1: only number of simulation
                2: simulation information
                3: simulation and episode information
                4: simulation, episode and round information
            - learn, a boolean to determine if agent learning is enabled.
        '''
        # Ask agent to make a decision
        try:
            action = self.agent.make_decision()
        except Exception as e:
            raise Exception('Oh oh', e)
        # Update records
        self.agent.actions.append(action)
        # Keeps the previous state before taking a step in the environment
        if verbose > 3:
            state = self.agent.states[-1]
            if hasattr(self.environment, 'state_visualizer'):
                state_visual = self.environment.state_visualizer(state)
            else:
                state_visual = str(state)
        # Runs the environment and obtains the next_state, reward, done, info
        result = self.environment.step(action)            
        next_state = self._clean_state(result[0])
        reward = result[1]
        done = result[2]
        # Prints info
        if verbose > 3:
            self._print_info(
                state=state, 
                state_visual=state_visual,
                action=action,
                next_state=next_state,
                reward=reward,
                done=done
            )
        # Agent learns
        if learn:
            self.agent.update(next_state, reward, done)
        # Update records
        self.agent.states.append(next_state)
        self.agent.rewards.append(reward)
        self.agent.dones.append(done)
        # Update round counter
        self.T += 1
        # Update environment "is-finished?"
        self.done = done

    def run(self, verbose:int=0, learn:bool=True) -> pd.DataFrame:
        '''
        Plays the specified number of rounds.
        Input:
            - verbose, to print information.
                0: no information
                1: only number of simulation
                2: simulation information
                3: simulation and episode information
                4: simulation, episode and round information
            - learn, a boolean to determine if agent learning is enabled.
        '''
        for round in range(self.num_rounds):
            print('>>', round, self.done)
            if not self.done:
                if verbose > 2:
                    print('\n' + '-'*10 + f'Round {round}' + '-'*10 + '\n')
                self.play_round(verbose=verbose, learn=learn)                
            else:
                print('Environment finished')
                break
            print(self.agent.states[-1])
        return self.to_pandas()

    def to_pandas(self) -> pd.DataFrame:
        '''
        Creates a pandas dataframe with the information from the current objects.
        Output:
            - pandas dataframe with the following variables:           
                Variables:
                    * episode: a unique identifier for the episode
                    * round: the round number
                    * action: the player's action
                    * reward: the player's reward
                    * done: whether the environment is active or not
                    * model: the model's name
                    * environment: the environment's name
        '''
        # Include las item in actions list
        self.agent.actions.append(np.nan)
        # Perform cleansing operations on registers
        states, actions, rewards, dones = self.cleanse_registers(
            [s for s in self.agent.states], 
            [a for a in self.agent.actions], 
            [r for r in self.agent.rewards], 
            [d for d in self.agent.dones]
        )
        data = {}
        data["episode"] = []
        data["round"] = []
        data["state"] = []
        data["action"] = []
        data["reward"] = []
        data["done"] = []
        for r in range(self.T):
            data["episode"].append(self.id)
            data["round"].append(r)
            data["state"].append(states[r])
            data["action"].append(actions[r])
            data["reward"].append(rewards[r])
            data["done"].append(dones[r])
        df = pd.DataFrame.from_dict(data)        
        df["model"] = self.model_name
        df["environment"] = self.env_name
        return df

    def cleanse_registers(
                self,
                states:List[any],
                actions:List[int],
                rewards:List[float],
                dones:List[bool]
            ) -> Tuple[List, List, List, List]:
        lengths = f'#states:{len(states)} -- #actions:{len(actions)} -- #rewards:{len(rewards)} -- #dones:{len(dones)}'
        n = len(states)
        assert(len(actions) == n), lengths
        assert(len(rewards) == n), lengths
        assert(len(dones) == n), lengths
        # Get indices from all nan in dones
        nan_dones = [i for i, d in enumerate(dones) if np.isnan(d)]
        # Check whether more than one episode lies in registers
        num_nan_dones = sum([1 for d in self.agent.dones if np.isnan(d)])
        if num_nan_dones > 1:
            # print('More than one episode in registers!')
            pass
        last_batch_init_index = nan_dones[-1]
        states = states[last_batch_init_index:]
        actions = actions[last_batch_init_index:]
        rewards = rewards[last_batch_init_index:]
        dones = dones[last_batch_init_index:]        
        assert(self.T == len(states)), f'T:{self.T}; len(dones):{len(dones)}\n{nan_dones}\n{dones}\n{self.agent.dones}'
        return states, actions, rewards, dones

    def reset(self) -> None:
        '''
        Reset the episode. This entails:
            reset the environment
            restart the agent 
        '''
        state = self.environment.reset()
        self.T = 1
        self.done = False
        state = self._clean_state(state)
        self.agent.restart()        
        if len(self.agent.states) == 1:
            if isinstance(self.agent.states, list):
                self.agent.states = [state]
            elif isinstance(self.agent.states, deque):
                self.agent.states = deque([state], maxlen=self.agent.max_len)
        else:
            self.agent.states.append(state)
        self.T = 1
        self.done = False

    def renderize(self, to_video:bool=False, file:str=None) -> None:
        '''
        Plays the specified number of rounds.
        '''
        if to_video:
            assert(file is not None), 'A folder name must be provided with the argument file='
            rm = self.environment.render_mode
            assert(rm == 'rgb_array'), f'To create video, environment render mode should be rgb_array, not {rm}'
            frame_list = []
        # Initialize img object
        img = plt.imshow(np.array([[0, 0], [0, 0]]))
        for round in range(self.num_rounds):
            if not self.done:
                im = self.environment.render()
                if isinstance(im, np.ndarray):
                    if to_video:
                        frame_list.append(im)
                    img.set_data(im)
                    plt.axis('off')
                    display(plt.gcf())
                sleep(self.sleep_time)
                clear_output(wait=True)
                self.play_round(verbose=0, learn=False)                
            else:
                clear_output(wait=True)
                im = self.environment.render()
                if isinstance(im, np.ndarray):
                    if to_video:
                        frame_list.append(im)
                    img.set_data(im)
                    plt.axis('off')
                    clear_output(wait=True)
                    display(plt.gcf())
                break
        if to_video:
            assert(len(frame_list) > 0), 'No frames saved. Check env.render() is providing np.arrays.'
            save_video(
                frames=frame_list,
                video_folder=file,
                fps=1/self.sleep_time
            )
   
    def _print_info(
                    self, 
                    state:any, 
                    state_visual:str, 
                    action:int, 
                    next_state:any, 
                    reward:float, 
                    done:bool
                ) -> None:
            """Prints the interaction"""
            if hasattr(self.environment, 'state_visualizer'):
                next_state_visual = self.environment.state_visualizer(next_state)
            else:
                next_state_visual = str(next_state_visual)
            if hasattr(self.environment, 'action_visualizer') and hasattr(self.agent, 'NN'):
                values_vector = self.agent.NN.values_vector(state)
                table = PrettyTable(['Action', 'Value'])
                action_names = [self.environment.action_visualizer(a) for a in range(self.agent.nA)]
                for action_name, value in zip(action_names, values_vector):
                    table.add_row([action_name, round(value, 2)])
            if hasattr(self.environment, 'action_visualizer'):
                action_visual = self.environment.action_visualizer(action)
            else:
                action_visual = str(action)
            print(f'\tThe state is => {state_visual}')
            print(f'\tThe values of the possible actions are:')
            print(table)
            print(f'\tAgent takes action => {action_visual}')
            print(f'\tThe state obtained is => {next_state_visual}')
            print(f'\tThe reward obtained is => {reward}')
            print(f'\tEnvironment is finished? => {done}')

    def simulate(
                self, 
                num_episodes:int=1, 
                file:str=None, 
                verbose:int=0, 
                learn:bool=True
            ) -> pd.DataFrame:
        '''
        Runs the specified number of episodes for the given number of rounds.
        Input:
            - num_episodes, int with the number of episodes.
            - file, string with the name of file to save the data on.
            - verbose, to print information.
                0: no information
                1: only number of simulation
                2: simulation information
                3: simulation and episode information
                4: simulation, episode and round information
            - learn, a boolean to determine if agent learning is enabled.
        Output:
            - Pandas dataframe with the following variables:

                Variables:
                    * id_sim: a unique identifier for the simulation
                    * round: the round number
                    * action: the player's action
                    * reward: the player's reward
                    * done: whether the environment is active or not
                    * model: the model's name
                    * environment: the environment's name
        '''
        # Create the list of dataframes
        data_frames = []
        # Run the number of episodes
        for ep in tqdm(range(num_episodes), desc='% of episodes run so far'):
            if verbose > 1:
                print('\n' + '='*10 + f'Episode {ep}' + '='*10 + '\n')
            # Reset the episode
            # lengths = f'#states:{len(self.agent.states)} -- #actions:{len(self.agent.actions)} -- #rewards:{len(self.agent.rewards)} -- #dones:{len(self.agent.dones)}'
            # print('//>', lengths)
            self.reset()
            self.id = ep
            # Run the episode
            # lengths = f'#states:{len(self.agent.states)} -- #actions:{len(self.agent.actions)} -- #rewards:{len(self.agent.rewards)} -- #dones:{len(self.agent.dones)}'
            # print('///>', lengths)
            df = self.run(
                verbose=verbose, 
                learn=learn
            )
            # Include episode in list of dataframes
            data_frames.append(df)
        # Concatenate dataframes
        data = pd.concat(data_frames, ignore_index=True)
        if file is not None:
            data.to_csv(file)
        return data
    

class Experiment :
    '''
    Compares given models on a number of measures.
    '''

    def __init__(self, \
                 environment:any, \
                 env_name:str, \
                 num_rounds:int, \
                 num_episodes:int, \
                 num_simulations:int, \
                 state_interpreter=id_state):
        '''
        Input:
            - environment, object with the environment on which to test the agents.
            - env_name, the environment's name.
            - num_rounds, int with the number of rounds.
            - num_episodes, int with the number of episodes.
            - num_simulations, int with the number of times the environment should be
                restarted and run the episodes again.
        '''
        self.environment = environment
        self.env_name = env_name
        self.num_rounds = num_rounds
        self.num_episodes = num_episodes
        self.num_simulations = num_simulations
        self.state_interpreter = state_interpreter
        self.data = None

    def run_experiment(self, \
                       agents:List[any], \
                       names:List[str], \
                       measures:List[str], \
                       learn:bool=True) -> None:
        '''
        For each agent, runs the simulation the stipulated number of times,
        obtains the data and shows the plots on the given measures.
        Input:
            - agents, list of agent objects.
            - names, list of names of the models implemented by the agents.
            - measures, list of measures, which could contain the following strings:
                * 'reward'
                * 'round_reward'
                * 'hist_reward'
            - learn, a boolean to enable learning.
        '''
        # Creates the list of dataframes
        data_frames = []
        # Run simulations
        for k in tqdm(range(self.num_simulations), desc='Running simulations'):
            # Reset all agents
            if learn:
                for agent in agents:
                    agent.reset()
            # Iterate over episodes
            for ep in tqdm(range(self.num_episodes), desc='\tRunning episodes', leave=False):
                # Initialize Episode
                sim_core = Episode(environment=self.environment, \
                                   env_name=self.env_name, \
                                   agent=None, \
                                   model_name=None,\
                                   num_rounds=self.num_rounds,\
                                   state_interpreter=self.state_interpreter)
                # Keep a unique id number for the episode
                sim_core.id = ep
                counter_agent = -1
                for agent in agents:
                    counter_agent += 1
                    # Copy Episode and place agent
                    sim = deepcopy(sim_core)
                    # Restart agent for a new episode
                    agent.restart()
                    sim.agent = agent
                    sim.agent.states.append(sim.initial_state)
                    sim.model_name = names[counter_agent]
                    # Run episode over agent
                    df = sim.run(verbose=False, learn=learn)
                    df["simulation"] = k
                    df["model"] = names[counter_agent]
                    data_frames.append(df)
        # Consolidate data
        data = pd.concat(data_frames, ignore_index=True)
        self.data = data
        # Create plots
        for m in measures:
            if m == 'reward':
                ax = Plot(data).plot_rewards(m)
            if m == 'round_reward':
                ax = Plot(data).plot_round_reward(m)
            if m == 'hist_reward':
                ax = Plot(data).plot_histogram_rewards(m)
            try:
                ax.set_title(m)
            except:
                pass
            plt.show()
        return agents

    def run_sweep1(self, \
                    agent:any, \
                    name:str, \
                    parameter:str, \
                    values:List[Union[int, float, str]], \
                    measures:List[str], \
                    learn:bool=True) -> None:
        '''
        For each agent, runs a parameter sweep the stipulated number
        of times, obtains the data and shows the plots on the given measures.
        Input:
            - agent, an object agent.
            - name, the name of the model implemented by the agent.
            - parameter, a string with the name of the parameter.
            - values, a list with the parameter's values.
            - measures, list of measures, which could contain the following strings:
                * 'reward'
                * 'round_reward'
            - learn, a boolean to enable learning.
        '''
        # Creates list of agents
        agents = []
        for value in values:
            agent_ = deepcopy(agent)
            instruction = f'agent_.{parameter} = {value}'
            exec(instruction)
            agents.append(agent_)
        # Creates list of names
        names = [f'({name}) {parameter}={value}' for value in values]
        # Run the experiment
        self.run_experiment(agents=agents,\
                            names=names,\
                            measures=[],\
                            learn=learn)            


class EnvfromGameAndPl2:
    '''
    Implementa un entorno a partir de un juego y del segundo jugador.
    '''
    def __init__(self, game:any, other_player:any):
        self.other_player = other_player
        self.initial_game = deepcopy(game)
        self.game = game
        self.state = self.game.estado_inicial
        self.list_acciones = None

    def reset(self):
        self.game = deepcopy(self.initial_game)
        self.state = self.game.estado_inicial
        self.other_player.reset()
        self.other_player.states.append(self.state)
        return self.state

    def render(self):
        self.game.render(self.state)

    def test_objetivo(self, state):
        if not self.game.es_terminal(state):
            return False
        else:
            player = self.game.player(state)
            return self.utilidad(state, player) > 0
        
    def acciones_aplicables(self, state):
        return self.game.acciones(state)

    def step(self, action):
        if self.list_acciones is not None:
            action_ = self.list_acciones[action] 
        else:
            action_ = action
        state = self.state
        playing = self.game.player(state)
        # print(f'player {playing} in state {state} makes move {action}')
        # First player made a move. Get new state, reward, done
        try:
            new_state = self.game.resultado(state, action_)
        except Exception as e:
            if action_ not in self.game.acciones(state):
                # Punish agent for playing an impossible action
                return state, -1000, True
            print(state)
            raise Exception(e)
        # self.game.render(new_state)
        # print(f'obtains {new_state}')
        reward = self.game.utilidad(new_state, playing)
        reward = reward if reward is not None else 0
        done = self.game.es_terminal(new_state)
        # If not done, second player makes a move
        if not done:
            playing = self.game.player(new_state)
            # Actualize second player with previous move
            self.other_player.states.append(new_state)
            if hasattr(self.other_player, 'choices'):
                possible_actions = self.game.acciones(new_state)
                self.other_player.choices = possible_actions
            # Second player makes a move
            other_player_action = self.other_player.make_decision()
            if self.other_player.debug:
                print(f'Negras mueven en {other_player_action}')
            # print(f'player {playing} in state {new_state} makes move {other_player_action}')
            # Get new state, reward, done
            new_state = self.game.resultado(new_state, other_player_action)
            # print(f'obtains {new_state}')
            reward = self.game.utilidad(new_state, playing)
            reward = reward if reward is not None else 0
            done = self.game.es_terminal(new_state)
        # Bookkeeping
        self.state = new_state
        return new_state, reward, done   
    



import os
import re

import streamlit as st
import polars as pl
import googlemaps

from chaparral.mapa.convenciones import POZO2CLUSTER, CLUSTER2MAPA


API_KEY = os.environ['GOOGLE_API_KEY']
OBJETOS_PATH = 'data/objetos_mapa/objetos_api.ndjson'

gmaps = googlemaps.Client(key=API_KEY)
objetos = pl.read_ndjson(OBJETOS_PATH)
bodega = objetos.filter(pl.col('codigo') == 'bodega_00')[0]


def clusters_por_quimico(quimico):
    dfi = inventario.filter(pl.col('quimico') == quimico)
    # return dfi
    puntos = []
    for row in dfi.iter_rows(named=True):
        cluster = None
        if row['punto'] in POZO2CLUSTER:
            # TODO: que hacer si el punto no esta en la lista?
            cluster = POZO2CLUSTER[row['punto']]
        elif row['punto'] in CLUSTER2MAPA.values():
            cluster = row['punto']
        else:
            st.write('missing cluster info for cluster:', row['punto'])
        if cluster:
            coords = objetos.filter(pl.col('nombre') == cluster)['coords'][0]
            puntos.append(dict(
                nombre=cluster,
                stock=row['stock'],
                lat=coords[0],
                lon=coords[1]
            ))
    return pl.DataFrame(puntos)


uploaded_file = st.file_uploader("Choose a file")
if uploaded_file is not None:
    raw_inventario = pl.read_csv(uploaded_file.read())
    inventario = load_inventario(raw_inventario)
else:
    st.stop()

quimicos = sorted(inventario.filter(pl.col('stock') > 0)['quimico'].unique())
quimico = st.selectbox('Seleccione Químico', quimicos)

data = clusters_por_quimico(quimico)
if data.is_empty():
    st.stop()

data = data.sort('stock', descending=False)
st.map(data.select(['lat', 'lon']))

for row in data.iter_rows(named=True):
    st.write('distancia to bodega')
    vecinos_bodega = bodega['vecinos'].to_list()[0]
    vecino = [v for v in vecinos_bodega if v['nombre'] == row['nombre']]
    if vecino:
        row['dst'] = vecino[0]['dist']
        row['tmp'] = vecino[0]['tiempo']
    st.write(row)




import json
import os

import googlemaps

from chaparral.mapa.convenciones import CLUSTER2MAPA

API_KEY = os.environ['GOOGLE_API_KEY']
gmaps = googlemaps.Client(key=API_KEY)

# Geocoding an address
ADDRESS = 'Cluster {cluster} Estación Chichimene, Acacías, Meta, Colombia'

geolocalizaciones = {}
for pozo in CLUSTER2MAPA.values():
    cluster = ADDRESS.format(cluster=pozo)
    geocode_result = gmaps.geocode(cluster)
    coords = geocode_result[0]['geometry']['location']
    geolocalizaciones[pozo] = coords

output_file_path = 'data/objetos_mapa/manual_geolocalizaciones.json'
json.dump(geolocalizaciones, open(output_file_path, 'w'), indent=4)

if __name__ == '__main__':
    pass



'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Define las constantes de información particular
de los clusters para usan en las varias clases.
-----------------------------------------------
'''

# Define el ruido a incluir en el tensor de estado
# para facilitar el aprendizaje de la red neuronal
RUIDO = 0
# Define qué tanto se multiplica el excedente en días
# de cada químico en cada pozo al finalizar el episodio
FACTOR_RECOMPENSA = 1e-3

# Define la recompensa por movimiento no permitido
RECOMPENSA_MOVIMIENTO_NO_PERMITIDO = -100

# Define el número de turnos para regresar a la bodega
HORAS_HABILES = 10
PENALIZACION_PASADAS_HORAS_HABILES = -100

# Define la capacidad del camion
NUM_BULKDRUMS = 8
NIVEL_MAXIMO_BULKDRUM = 250

# Define qué cantidad de días en excedente debe haber
# de cada químico en cada pozo para evitar penalización
VALOR_CRITICO_CARGA = 8 # usado para el cálculo de cuanto cargar en pozo
VALOR_CRITICO_TOLERANCIA = 0.75 * VALOR_CRITICO_CARGA # usado para el cálculo de la recompensa

# Define el tiempo de preparación para cargar un químico en un pozo
TIEMPO_DESCARGA = 1 / 60 * 25

# Define el escalamiento de la velocidad de consumo de un químico
# para pasar de la info suministrada (Gl/día) a la usada en el
# entorno (Gl/h)
ESCALA_VELOCIDAD_CONSUMO = 1 / 24

# Define el escalamiento de la velocidad del camión para pasar
# de la info suministrada (????/milisegundos) a la usada en el
# entorno (kh/h)
ESCALA_VELOCIDAD_CAMION = 1

# Define el escalamiento del tiempo de ruta
ESCALA_TIEMPO_RUTA = 1 / (60 * 60)

# Define el escalamiento para el render
FACTOR_RENDER_X = 2
FACTOR_RENDER_Y = 15
MEDIA_X = 3.9313048491215072
DESVEST_X = 0.019056068943662685
MEDIA_Y = -73.69699527751696
DESVEST_Y = 0.026686490709401194


import numpy as np
import streamlit as st
import polars as pl
import plotly.express as px

path = 'data/Inventario_Pozos.csv'
df = pl.read_csv(path)

estaciones = df['Cod.'].unique().to_list()
productos = df['Producto'].unique().to_list()

coords = df.select('Cod.').unique(). \
            with_columns([
                pl.lit(np.random.uniform(3.8925,  3.8955, len(estaciones))).alias('lat'),
                pl.lit(np.random.uniform(73.6645, 73.6675, len(estaciones))).alias('lon')
            ])

"""
⁦3°56'07"N⁩ ⁦73°40'33"W⁩
⁦3°56'04"N⁩ ⁦73°40'49"W⁩
⁦3°55'57"N⁩ ⁦73°41'14"W⁩
"""

stock_product = df. \
         select(['Producto', 'Stock en días']). \
         group_by('Producto'). \
         first(). \
         sort('Stock en días')

stock_pozos = df. \
         select(['Cod.', 'Stock en días']). \
         group_by('Cod.'). \
         first(). \
         sort('Stock en días').\
         with_columns(('Pozo_' + pl.col('Cod.').cast(str)).alias('Pozo'))

total_zero_products = stock_product.filter(pl.col('Stock en días') == 0).shape[0]
total_zero_pozos = stock_pozos.filter(pl.col('Stock en días') == 0).shape[0]

fig_low_product = px.bar(stock_product.head(10),
                         y='Producto',
                         x='Stock en días',
                         title='Productos Bajos en Stock'
                         )

fig_low_pozos = px.bar(stock_pozos.head(10),
                       y='Pozo',
                       x='Stock en días',
                       title='Pozos Bajos en Stock'
                       )

st.header('Proyecto Dist XYZ')
st.plotly_chart(fig_low_product)
st.plotly_chart(fig_low_pozos)

st.write(f'Total de productos con 0 stock: {total_zero_products}, Total de pozos con 0 stock: {total_zero_pozos}')

tab1, tab2 = st.tabs(['Todos', '0 Stock'])
with tab1:
    st.map(coords, size=1)
with tab2:
    st.map(coords.filter(stock_pozos['Stock en días'] == 0), size=1)




'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Class to XXXX.
-----------------------------------------------
'''


NOMBRES_CLUSTERS = ['bodega_00', 'A', 'B', 'C']

DATOS_RUTAS = {
	'Bodega <-> A':{
        'longitud':2,
        'velocidad_promedio':2
    },
	'A <-> Bodega':{
        'longitud':2,
        'velocidad_promedio':2
    },
#
	'Bodega <-> C':{
        'longitud':4,
        'velocidad_promedio':1
    },
	'C <-> Bodega':{
        'longitud':4,
        'velocidad_promedio':1
    },
#
	'A <-> B':{
        'longitud':2,
        'velocidad_promedio':2
    },
	'B <-> A':{
        'longitud':2,
        'velocidad_promedio':2
    },
#
	'B <-> C':{
        'longitud':2,
        'velocidad_promedio':2
    },
	'C <-> B':{
        'longitud':2,
        'velocidad_promedio':2
    },
}

#C = Cluster(
#    nombre='C',
#    activo=False,
#    quimicos=[xyz],
#    niveles_actuales=[0],
#    velocidades_consumo=[0],
#    coordenadas=(-236.3568539525139, 115.90576491902122),
#    vecinos=['bodega_00', 'B'],
#)



'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Clase para correr el campo de pozos con sus
clusters y camiones como un espacio de estados.
-----------------------------------------------
'''
from chaparral.modelos.entorno import CampoPozos
from chaparral.modelos.clases import Cluster, Camion, Ruta, PseudoRutaCamionEstacionario, Quimico
from chaparral.modelos.constantes import DATOS_RUTAS
from chaparral.modelos.busqueda import Nodo, solucion
from typing import List, Dict, Optional, Tuple, Union
import torch
import warnings 
import numpy as np
from itertools import product
from copy import deepcopy

class EspacioEstados :
	"""Crea el campo de pozos como un espacio de estados"""
	def __init__(self, campo:CampoPozos):
		self.campo = campo
		self.estado_inicial = deepcopy(campo.obtener_estado())

	def acciones_aplicables(
				self, 
				estado:torch.tensor
			) -> List[str]:
		'''
		Devuelve la lista de clusters accessibles
		desde el cluster en que está el camión.
		Input: estado ///// NO SE USA, PERO LO REQUIERE EL METODO DE BUSQUEDA
		Output: lista de acciones correspondientes a las
				combinaciones de clusters accesibles desde 
		 		el lugar donde está cada camión.
		'''
		# Actualiza atributos de acuerdo al estado recibido
		info_estado = self._tensor_a_atributos(estado)
		self._actualiza_desde_info(info_estado)
		# Encontramos los índices de los clusters donde están los camiones
		lista_indices_clusters = [self.campo.dict_clusters[l] for l in self.campo.loc_camiones]
		# Encontramos los clusters veciones de cada locación
		lista_clusters_vecinos = [self.campo.clusters[indice_cluster].vecinos for indice_cluster in lista_indices_clusters]
		lista_clusters_vecinos = [[self.campo.dict_clusters[cluster] for cluster in vecinos] for vecinos in lista_clusters_vecinos]
		# Encontramos las combinaciones posibles
		combinaciones_movimientos = product(*lista_clusters_vecinos)
		combinaciones_movimientos = [list(comb) for comb in combinaciones_movimientos]
		# Encontramos el índice de cada combinación
		lista_acciones = [self.campo.movimientos_a_accion(lista_movimientos) for lista_movimientos in combinaciones_movimientos]
		return lista_acciones

	def transicion(
				self, 
				estado:torch.tensor, 
				accion:int
			) -> torch.tensor:
		'''
		Devuelve el estado resultado de realizar la acción
		en el estado dado
		Input: estado, tensor
		       accion, que es una ciudad
		Output: estado, tensor
		'''
		# Actualiza atributos de acuerdo al estado recibido
		info_estado = self._tensor_a_atributos(estado)
		self._actualiza_desde_info(info_estado)
		# Actualiza bookkeeping de localización de camiones
		locaciones_antes = self.campo.loc_camiones.copy()
		# Hace el step de acuerdo a la acción
		nuevo_estado, recompensa, done = self.campo.step(accion)
		info_nuevo_estado = self._tensor_a_atributos(nuevo_estado)
		# Actualiza bookkeeping de localización de camiones
		locaciones_despues = self.campo.loc_camiones.copy()
		distancias = self._obtiene_distancias_recorridas(
			locaciones_antes=locaciones_antes, 
			locaciones_despues=locaciones_despues,
			unidades_temporales=self.campo._unidades_temporales
		)
		# Actualizamos bookkeeping
		self._check_estado = str(estado)
		self._fake_costo = sum(distancias)
		self._fake_test_objetivo = done
		return nuevo_estado
	
	def _tensor_a_atributos(
				self, 
				estado:torch.tensor
			) -> Dict[str, Dict[str,any]]:
		'''
		Devuelve una lista con los atributos del entorno
		a partir de su representación tensorial.
		'''
		# Vemos el tensor como lista
		estado_como_lista = estado.tolist()
		# Inicializamos registros
		valores_clusters = {cluster.nombre:list() for cluster in self.campo.clusters}
		valores_camiones = {camion.nombre:list() for camion in self.campo.camiones}
		indice_inicial = 0
		#-----------------------------------------------
		# 1. Encontramos la información de cada cluster
		#-----------------------------------------------
		for i, cluster in enumerate(self.campo.clusters):
			# Recorre cada cluster
			# Cada cluster está representado por dos valores para cada químico (cantidad, velocidad)
			# y un valor booleano de si está activo o no
			indice_final = (i + 1) * (2 * len(cluster.quimicos) + 1) 
			lista_estado_cluster = estado_como_lista[indice_inicial:indice_final]
			cantidades, velocidades, activo = self._encuentra_info_cluster(
				lista_estado_cluster,
				cluster.quimicos
			)
			# Guardamos toda la información del cluster
			valores_clusters[cluster.nombre] = {
				'cantidades':cantidades,
				'velocidades':velocidades,
				'activo':activo 
			}
			indice_inicial = indice_final
		#-----------------------------------------------
		# 2. Encontramos la información de los camiones
		#-----------------------------------------------
		for i, camion in enumerate(self.campo.camiones):
			# Recorre cada camion
			# Cada camión está representado por la carga de cada químico y un one-hot de los clusters
			indice_final += camion.num_quimicos + self.campo.num_clusters
			lista_estado_camion = estado_como_lista[indice_inicial:indice_final]
			carga, nombre_cluster = self._encuentra_info_camion(
				lista_estado_camion,
				camion.num_quimicos
			)
			# Guardamos la información del camion
			valores_camiones[camion.nombre] = {
				'carga':carga,
				'ubicacion':nombre_cluster
			}
			indice_inicial = indice_final
		return {'clusters':valores_clusters, 'camiones':valores_camiones}

	def _encuentra_info_cluster(
			self,
			lista_estado_cluster:List[float],
			quimicos:List[Quimico]
		) -> Tuple[List[float], List[float], bool]:
		'''Encuentra la información del cluster en la lista proveniente del tensor'''
		cantidades = list()
		velocidades = list()
		for j, quimico in enumerate(quimicos):
			# Recorremos cada químico en el cluster
			# El primer argumento del químico es su cantidad
			cantidades.append(lista_estado_cluster[j])
			# El segundo argumento es su velocidad de consumo
			velocidades.append(lista_estado_cluster[j + 1])
		# El argumento final es si el pozo está activo o no
		activo = lista_estado_cluster[-1]
		return cantidades, velocidades, activo
	
	def _encuentra_info_camion(
			self,
			lista_estado_camion:List[float],
			num_quimicos:int
		) -> Tuple[float, str]:
		'''Encuentra la información del camión en la lista proveniente del tensor'''
		carga = lista_estado_camion[:num_quimicos]
		indice_cluster = np.argmax(lista_estado_camion[num_quimicos:])
		nombre_cluster = self.campo.clusters[indice_cluster].nombre
		return carga, nombre_cluster
	
	def _actualiza_desde_info(
			self,
			info_estado:Tuple[Dict, Dict]
		) -> None:
		"""Actualiza los atributos del entorno de acuerdo a info"""
		for cluster in self.campo.clusters:
			cluster.niveles_actuales = info_estado['clusters'][cluster.nombre]['cantidades']
		loc_camiones = list()
		for camion in self.campo.camiones:
			camion.carga_cantidades = info_estado['camiones'][camion.nombre]['carga']
			loc_camiones.append(info_estado['camiones'][camion.nombre]['ubicacion'])
		self.campo.loc_camiones = loc_camiones

	def _obtiene_distancias_recorridas(
				self,
				locaciones_antes:List[int], 
				locaciones_despues:List[int],
				unidades_temporales:float
			) -> List[float]:
		"""Encuentra las distancias recorridas por los camiones"""
		nombres_rutas = zip(locaciones_antes, locaciones_despues)
		nombres_rutas = [f'{par[0]} <-> {par[1]}' for par in nombres_rutas]
		velocidades = [DATOS_RUTAS[ruta]['velocidad_promedio'] for ruta in nombres_rutas]
		distancias = [velocidad * unidades_temporales for velocidad in velocidades]
		return distancias

	def test_objetivo(self, nodo:Nodo):
		# Devuelve True/False dependiendo si el estado
		# resuelve el problema
		# Input: estado, que es una ciudad
		# Output: True/False
		# Actualiza atributos de acuerdo al estado recibido
		estado = nodo.estado
		info_estado = self._tensor_a_atributos(estado)
		self._actualiza_desde_info(info_estado)
		primer_check = not self.campo.comienzo
		segundo_check = np.all([l == 'bodega_00' for l in self.campo.loc_camiones])
		camino = solucion(nodo)
		tercer_check = np.all([cluster in camino for cluster in list(range(self.campo.num_clusters))])
		if np.all([primer_check, segundo_check, tercer_check]):
			return True
		return False

	def costo(self, estado, accion):
		# Devuelve la distancia desde estado hasta accion (que es una ciudad)
		# Input: 
		#        estado, que es una ciudad
		#        accion, que es una ciudad
		# Output: distancia de acuerdo al diccionario rutas
		try:
			if str(estado) == self._check_estado:
				return self._fake_costo
		except:
			warnings.warn(f'costo() requiere haber corrido transicion() antes.') 
			return 0

	def codigo(self, estado):
		return str(estado)

	def render(self):
		return


# copyright ardillabyte.com 2023

import itertools
from tqdm import tqdm
from dataclasses import dataclass
from typing import Tuple, List, Optional, Sequence, Dict
from enum import Enum
from pathlib import Path
import jsonlines

import numpy as np
from numpy import ndarray
import cv2 as cv
from cv2 import Mat

GREEN_LOW = np.array((42, 193, 243))
GREEN_HIGH = np.array((49, 199, 253))

ERODE_SIZE = 5
DILATE_SIZE = 10
SHAPE = cv.MORPH_ELLIPSE


def get_structuring_element(size: int, shape=SHAPE):
    return cv.getStructuringElement(shape,
                                    (2 * size + 1, 2 * size + 1),
                                    (size, size))


ERODE_ELEMENT = get_structuring_element(ERODE_SIZE)
DILATE_ELEMENT = get_structuring_element(DILATE_SIZE)


class ObjetoTipo(Enum):
    CLUSTER = 'cluster'
    VIA_PAVIMENTADA = 'pavimentada'
    VIA_PAVIMENTADA_ANGOSTA = 'angosta'
    VIA_SIN_PAVIMENTAR = 'sin_pavimentar'
    BODEGA = 'bodega'

    def __repr__(self):
        return self.name


@dataclass
class ObjetoMapa:
    """Objeto en el mapa."""
    nombre: str
    tipo: str
    contorno: List[Tuple[int, int]]
    vecinos: list = None

    def __repr__(self):
        return self.nombre


@dataclass
class Arista:
    nombre: str
    dist: float

    def __repr__(self):
        return f'{self.nombre} {self.dist}'


class Mapa:
    def __init__(self):
        self.objetos: Optional[List[ObjetoMapa]] = None
        self.mapa = None

    def cargar_objetos_en_imagen(self, img_path: Path, tipo: str, factor=4):
        tipo = [o for o in ObjetoTipo if tipo == o.value]
        if tipo:
            tipo = tipo[0]
        else:
            raise ValueError(f'Tipo {tipo} not recognized.')
        objetos = self._encontrar_objetos(img_path, factor)
        objetos_mapa = []
        for i, objeto in enumerate(objetos):
            objetos_mapa.append(
                ObjetoMapa(
                    nombre=f'{tipo.value}_{i:02d}',
                    tipo=tipo.name,
                    contorno=[self._encontrar_centro(objeto)]
                    if tipo.value in ['cluster', 'bodega']
                    else objeto.tolist(),
                )
            )
        if self.objetos is None:
            self.objetos = objetos_mapa
        else:
            self.objetos.extend(objetos_mapa)

    def encontrar_vecinos(self, limite: int = 30):
        for a in tqdm(self.objetos, desc='Encontrando Vecinos'):
            if a.vecinos is None:
                a.vecinos = []
            for b in tqdm(self.objetos, desc=f'encontrando vecinos de {a.nombre}'):
                if a != b:
                    arista_minima = self._encontrar_distancia_minima(a, b)
                    if arista_minima.dist < limite:
                        a.vecinos.append(arista_minima.nombre)

    def to_json(self, nombre_archivo: str):
        with jsonlines.open(nombre_archivo, 'w') as writer:
            for objeto in self.objetos:
                writer.write([objeto.__dict__])

    def _encontrar_objetos(self, img_path: Path, factor: int) -> Sequence[Mat]:
        if not img_path.exists():
            raise ValueError(f'No se encuentra el archivo {img_path}')
        img = cv.imread(str(img_path))
        if img is not None:
            hsv = cv.cvtColor(img, cv.COLOR_BGR2HSV)
            mask = cv.inRange(hsv, GREEN_LOW, GREEN_HIGH)
            mask = cv.erode(mask, ERODE_ELEMENT)
            mask = cv.dilate(mask, DILATE_ELEMENT)
            mask = cv.resize(mask, (int(mask.shape[0] / factor), int(mask.shape[1] / factor)))
            if self.mapa is None:
                self.mapa = mask
            else:
                self.mapa += mask
            contours, _ = cv.findContours(mask, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)
            return contours

    def _encontrar_distancia_minima(self, c1: ObjetoMapa, c2: ObjetoMapa) -> Arista:
        puntos = list(itertools.product(c1.contorno, c2.contorno))
        distancias = [
            Arista(
                nombre=c2.nombre,
                dist=self._medir_distancia(a, b)
            )
            for a, b in puntos]
        return min(distancias, key=lambda x: x.dist)

    @staticmethod
    def _medir_distancia(a: Tuple[int, int], b: Tuple[int, int]) -> float:
        if not isinstance(a, ndarray):
            a = np.array(a)
        if not isinstance(b, ndarray):
            b = np.array(b)
        return np.sqrt(np.sum((a - b) ** 2))

    @staticmethod
    def _encontrar_centro(objeto) -> Tuple[int, int]:
        x, y = objeto.mean(axis=0)[0]
        return int(x), int(y)



import json
import os
from tqdm import tqdm

import polars as pl
import googlemaps

from chaparral.mapa.convenciones import CLUSTER2MAPA

API_KEY = os.environ['GOOGLE_API_KEY']
gmaps = googlemaps.Client(key=API_KEY)

geocodes_path = 'data/objetos_mapa/manual_geolocalizaciones.json'
geocodes = json.load(open(geocodes_path, 'r'))

vecinos_path = 'data/objetos_mapa/objetos.ndjson'
df = pl.read_ndjson(vecinos_path)


clusters = [objeto for objeto in df['nombre'].unique() if 'cluster' in objeto]
clusters.append('bodega_00')

aristas = []
for cluster in tqdm(clusters, desc='Processing vecinos'):
    cluster_nombre = CLUSTER2MAPA.get(cluster.upper(), None)
    vecinos = get_vecinos(cluster)
    if not cluster_nombre:
        print(f'No encuentro el nomber del cluster {cluster}')
        continue
    vecinos_nombres = [CLUSTER2MAPA.get(v.upper()) for v in vecinos]
    for vecino in vecinos_nombres:
        request = gmaps.distance_matrix(
            geocodes[cluster_nombre],
            geocodes[vecino],
            mode='driving',
            units='metric'
        )
        row = dict(
            src=cluster_nombre,
            trg=vecino,
            dist=request['rows'][0]['elements'][0]['distance']['value'],
            tiempo=request['rows'][0]['elements'][0]['duration']['value'],
        )
        aristas.append(row)

aristas = pl.DataFrame(aristas)
aristas.write_ndjson('data/objetos_mapa/aristas.ndjson')
print(aristas)



'''
© Copyright ArdillaByte Inc. 2023

------------------------------------
Module with the state interpreters.
------------------------------------
'''
import numpy as np
import torch
from typing import Union, Tuple

def id_state(state):
    '''
    Default interpreter: do nothing.
    '''
    return state

def gym_interpreter1(state):
    '''
    Cleans the state and get only the state space.
    When states come from gymnasium, they contain 
    additional info besides the state space.
    '''
    if isinstance(state, tuple):
        if isinstance(state[1], dict):
            state = state[0]
        else:
            state = state
    else:
        state = state
    return state

def gym_interpreter2(state):
    '''
    Cleans the state and get only the state space.
    When states come from gymnasium, they contain 
    additional info besides the state space.
    Returns a pytorch tensor.
    '''
    if isinstance(state, tuple):
        if isinstance(state[1], dict):
            state = state[0]
        else:
            state = state
    else:
        state = state
    return torch.tensor(state, dtype=torch.float32)

def triqui_interpreter(state):
    '''
    Interprets triqui state as a ravel index.
    '''
    shape = tuple([3]*9)
    comps = tuple(state.flatten().tolist()[0])
    indice = np.ravel_multi_index(comps, shape)
    return indice

def blackjack_interpreter(state):
    '''
    Interprets blackjack state as a ravel index.
    The observation consists of a 3-tuple containing: 
        the player’s current sum, 
        the value of the dealer’s one showing card (1-10 where 1 is ace), 
        and whether the player holds a usable ace (0 or 1).
    The observation is returned as (int(), int(), int()).
    We transform the 3-tuple into an index using numpy.ravel_multi_index
    '''
    shape = (32, 11, 2)
    # Ignore extra info from state
    if len(state) == 2:
        state_ = state[0] 
    else:
        state_ = state
    indice = np.ravel_multi_index(state_, shape)
    return indice

def PitL_nS_interpreter(state):
    '''
    Interprets PitLaberynth state as a ravel index.
    '''
    shape = (state.shape[1], state.shape[2])
    comps = np.where(state == 1)
    to_ravel = [(comps[1][i],comps[2][i]) for i in range(len(comps[0]))]
    ravels = [np.ravel_multi_index(mi, shape) for mi in to_ravel]
    n = np.product(shape)
    n_shape = (n, n, n, n)
    return np.ravel_multi_index(ravels, n_shape)

def PitL_cs_interpreter(state):
    '''
    Interprets PitLaberynth state as a triple.
    '''
    shape = (state.shape[1], state.shape[2])
    comps = np.where(state == 1)
    to_ravel = [(comps[1][i],comps[2][i]) for i in range(len(comps[0]))]
    ravels = [np.ravel_multi_index(mi, shape) for mi in to_ravel]
    return tuple(ravels)

def PitL_vector_interpreter(state):
    '''
    Interprets PitLaberynth state as a single vector
    '''
    shape = np.product(state.shape)
    return state.reshape(1, shape)

def PitL_CNN_interpreter(state:np.ndarray) -> torch.tensor:
    '''
    Interprets PitLaberynth state as an image
    with 4 channels and returns a pytorch tensor
    with only 1 channel
    '''
    projection = np.zeros(state[0].shape)
    # Projecting layers
    for i, layer in enumerate(state):
        projection += layer*(i+1)
    # Converting to tensor
    projection = torch.tensor(projection, dtype=torch.float32)
    # Unsqeeze for convolutional network with 1 channel
    projection = projection.unsqueeze(dim=0)
    return projection

def GridW_CNN_interpreter(state:Union[int, Tuple]) -> torch.tensor:
    '''
    Interprets GridWorld state as an image
    with 1 channel and returns a pytorch tensor
    '''
    # Initialize array
    state_ = np.zeros((4,4))
    if isinstance(state, int) or isinstance(state, np.int64):
        # Find coordinates
        Y, X = np.unravel_index(state, (4,4))
    elif isinstance(state, Tuple):
        X, Y = state
        X, Y = int(X), 3 - int(Y)
    else:
        raise Exception(f'Unreadable state {state}. Must be of type int or tuple, got {type(state)}')
    # See as matrix indices
    row = 3 - Y
    col = X
    state_[row,col] = 1
    # Convert to tensor
    state_ = torch.tensor(state_, 
                          dtype=torch.float32)
    # Unsqeeze for convolutional network with 1 channel
    state_ = state_.unsqueeze(dim=0)
    return state_



'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Class to XXXX.
-----------------------------------------------
'''
import numpy as np
from typing import List, Dict, Optional, Tuple, Union
from copy import deepcopy
import torch
import torch.nn.functional as F
import pandas as pd

from chaparral.modelos.constantes import NIVEL_MAXIMO_BULKDRUM


def a_tabla(cadena):
    c = '<table>\n<tbody>\n'
    c += cadena
    c += '</tbody>\n</table>'
    return c

class Quimico:
	'''
	Objeto quimico

	Atributos:
		- nombre
		- unidades de medida
	'''
	def __init__(
				self, 
				nombre:str, 
				unidades:str
			) -> None:
		self.nombre = nombre
		self.unidades = unidades

	def __str__(self) -> str:
		return f'{self.nombre} (unidades:{self.unidades})'

	def in_carga(
				self, 
				carga_quimicos:List[any]
			) -> Union[None, int]:
		'''Determina si self está en una lista y en caso positivo devuelve el índice en la lista'''
		nombres_quimicos = [quimico.nombre for quimico in carga_quimicos]
		if self.nombre in nombres_quimicos:
			return nombres_quimicos.index(self.nombre)
		else:
			return None


class QuimicoEn(Quimico):

	def __init__(
				self, 
				nombre: str, 
				unidades: str,
				nivel_actual: float,
				velocidad_consumo: float
			) -> None:
		super().__init__(nombre, unidades)
		self.nivel_actual = nivel_actual
		self.nivel_inicial = nivel_actual
		self.velocidad_consumo = velocidad_consumo

	def __str__(self) -> str:
		''' Genera una cadena en formato html con el nombre del químico, su nivel y velocidad'''
		cadena = f'<td> {self.nombre} </td> '
		cadena += f'<td> {round(self.nivel_actual, 2)} </td> '
		if self.velocidad_consumo is not None:
			cadena += f'<td> {round(self.velocidad_consumo, 2)} </td>'
		return cadena

	def actualizar(
				self, 
				unidades_temporales:float
			) -> None:
		self.nivel_actual -= self.velocidad_consumo * unidades_temporales
		self.nivel_actual = max(0, self.nivel_actual)
	
	def a_tensor(self, en_camion=False) -> torch.tensor:
		if en_camion or self.velocidad_consumo is None:
			return torch.tensor([self.nivel_actual], dtype=torch.float32)			
		else:
			return torch.tensor([self.nivel_actual, self.velocidad_consumo], dtype=torch.float32)
	

class Pozo:
	'''
	Objeto pozo

	Atributos:
		- nombre
		- si está activo o no
		- lista de químicos en pozo
	'''
	def __init__(
				self,
				nombre:str,
				quimicos:List[QuimicoEn]
			) -> None:
		self.nombre = nombre
		self.quimicos = quimicos	
		self.num_quimicos = len(quimicos)

	def __str__(self) -> str:
		cadena = f'\n<tr>\n\t<td> {self.nombre} </td> '
		inicial = True
		for quimico in self.quimicos:
			if inicial:
				cadena += str(quimico) + '\n</tr> '	
				inicial = False
			else:
				cadena += ' \n<tr>\n\t<td>&nbsp;</td> ' + str(quimico) + '\n</tr>'	
		return cadena

	def actualizar(
				self,
				unidades_temporales:float
			) -> None:
		for quimico in self.quimicos:
			quimico.actualizar(unidades_temporales)

	def a_tensor(self) -> torch.tensor:
		tensores = [quimico.a_tensor() for quimico in self.quimicos]
		return torch.cat(tensores)


class Cluster:
	'''
	Objeto cluster

	Atributos:
		- nombre
		- lista de pozos
		- coordenadas del cluster
	'''
	def __init__(
				self, 
				nombre:str, 
				pozos:List[str],
				coordenadas:Tuple, 
			) -> None:
		self.nombre = nombre
		# -----------------------------------
		# Los pozos del cluster
		# -----------------------------------
		self.pozos = pozos
		self.num_pozos = len(self.pozos)
		# -----------------------------------
		# Atributos geográficos
		# -----------------------------------
		self.coordenadas = coordenadas # Coordenadas geográficas del cluster
		
	def actualizar(
			self, 
			unidades_temporales:int
			) -> None:
		'''
		Actualiza el cluster simulando que transcurren las unidades temporales dadas

		Input:
			- unidades_temporales (int), la cantidad de unidades de tiempo que transcurren.
		'''
		for pozo in self.pozos:
			pozo.actualizar(unidades_temporales)
	
	def __str__(self) -> str:
		'''Genera una cadena en formato html con el nombre de cada químico y su cantidad actual'''
		cadena = '<tr>\n\t<td>Cluster</td> \n</tr>'
		cadena += f'\n<tr>\n\t<td> {self.nombre} </td> '
		inicial = True
		for pozo in self.pozos:
			if inicial:
				str_pozo = '<tr>\n\t<td>Pozo</td> <td>Químico</td> <td>Cantidad</td> <td>Vel.Cons</td> \n</tr>' + str(pozo)
				cadena += ' <td> ' + a_tabla(str_pozo) + ' </td> \n</tr> '	
				inicial = False
			else:
				str_pozo = '<tr>\n\t<td>Pozo</td> <td>Químico</td> <td>Cantidad</td> <td>Vel.Cons</td> \n</tr>' + str(pozo)
				cadena += ' <tr>\n\t<td>&nbsp;</td> ' + ' <td> ' + a_tabla(str_pozo) + ' </td> \n</tr> '	
		return a_tabla(cadena)
	
	def cluster_a_tensor(self) -> torch.tensor:
		'''Genera un tensor plano donde, por cada químico, incluye 
			el nivel actual,
			la velocidad de consumo.
		'''
		tensores = [pozo.a_tensor() for pozo in self.pozos]
		assert(len(tensores) > 0), f'¡Error!: ¡Cluster {self.nombre} no tiene pozos!'
		return torch.cat(tensores)


class Ruta:
	'''
	Objeto que representa la ruta que está siguiendo 
	un camión para ir de un cluster a otro.

	Atributos:
		- nombre del cluster inicial
		- nombre del cluster objetivo
		- tiempo de tránsito de un camión en kilómetros por unidad temporal
	'''
	def __init__(
				self, 
				inicial:Cluster, 
				objetivo:Cluster,
				tiempo:float,
			) -> None:
		self.inicial = inicial
		self.objetivo = objetivo
		self.nombre = f'{self.inicial.nombre} <-> {self.objetivo.nombre}'
		self.tiempo = tiempo
		self.porcentaje_recorrido = 0

	def __str__(self) -> str:
		cadena = f'Ruta {self.nombre} por {round(self.longitud,2)}km a {self.velocidad_promedio}km/hora'
		cadena += '\n\t' + f'El camion ha recorrido el {round(self.porcentaje_recorrido, 2)}%'
		return cadena

	def actualizar(self, unidades_temporales:float) -> None:
		'''
		Mueve el camión por la ruta de acuerdo a la
		cantidad de unidades_temporales dada 
		'''
		tiempo_faltante = self.obtener_unidades_temporales_faltantes()
		if  unidades_temporales >= tiempo_faltante:
			# El camión llega al final de la ruta
			self.porcentaje_recorrido = 100
		else:
			tiempo_recorrido = self.tiempo - tiempo_faltante
			self.porcentaje_recorrido = (tiempo_recorrido + unidades_temporales) / self.tiempo * 100

	def fin_recorrido(self) -> bool:
		# Determina si el camino fue recorrido
		return self.porcentaje_recorrido == 100
	
	def obtener_unidades_temporales_faltantes(self) -> float:
		'''
		Determina cuántas unidades temporales faltan para terminar la ruta
		'''
		tiempo_faltante = (100 - self.porcentaje_recorrido) / 100 * self.tiempo
		return tiempo_faltante


class PseudoRutaCamionEstacionario(Ruta):
	'''
	Clase que simula una ruta pero establece
	los métodos de ruta para cuando el camión
	está estacionario.
	'''
	def __init__(
				self, 
				inicial: Cluster, 
				objetivo: Cluster, 
				tiempo: float
			) -> None:
		super().__init__(inicial, objetivo, tiempo)

	def __str__(self) -> str:
		return f'Camión estacionado en {self.inicial.nombre}'
	
	def actualizar(self, unidades_temporales: float) -> None:
		return None
	
	def fin_recorrido(self) -> bool:
		return True
	
	def obtener_unidades_temporales_faltantes(self) -> float:
		return np.Inf
	

class BulkDrum:
	'''
	Objeto bulkdrum --- contenedor de químico en el camión
	'''
	def __init__(
				self, 
				nombre:str,
				nombre_quimico:str,
				nivel_actual:float,
			) -> None:
		self.nombre = nombre
		self.quimico = Quimico(
			nombre=nombre_quimico,
			unidades='Gl'
		)
		self.nivel_actual = nivel_actual
		self.nivel_inicial = nivel_actual
		self.nivel_maximo = NIVEL_MAXIMO_BULKDRUM
	
	def __str__(self) -> str:
		''' Genera una cadena en formato html con el nombre del químico, su nivel y velocidad'''
		cadena = f'<td> {self.quimico.nombre} </td> '
		cadena += f'<td> {round(self.nivel_actual, 2)} </td> '
		return cadena

	def llenar(self, cantidad_a_cargar:float) -> None:
		"""Intenta llenar el blulkdrum con la cantidad dada"""
		if self.nivel_actual + cantidad_a_cargar <= self.nivel_maximo:
			self.nivel_actual += cantidad_a_cargar
		else:
			raise Exception(f'Error de llenado bulkdrum {self.nombre} a nivel {round(self.nivel_actual, 2)}. Cantidad {round(cantidad_a_cargar, 2)} supera límite máximo')

	def llenar_suavemente(
				self, 
				cantidad_a_cargar:float
			) -> float:
		"""Llena el bulkdrum y devuelve la cantidad que sobra"""
		if self.nivel_actual + cantidad_a_cargar <= self.nivel_maximo:
			self.nivel_actual += cantidad_a_cargar
			return 0
		else:
			cantidad_faltante = self.nivel_actual + cantidad_a_cargar - self.nivel_maximo
			self.nivel_actual = self.nivel_maximo
			return cantidad_faltante

	def descargar(
				self, 
				cantidad_a_descargar:float
			) -> float:
		"""Descarga el bulkdrum y devuelve la cantidad que falta por descargar"""
		if self.nivel_actual > cantidad_a_descargar:
			self.nivel_actual -= cantidad_a_descargar
			return 0
		else:
			aux = self.nivel_actual
			self.nivel_actual = 0
			return cantidad_a_descargar - aux

	def con_capacidad(self) -> bool:
		return self.nivel_actual < self.nivel_maximo

	def a_tensor(self) -> torch.tensor:
		return torch.tensor([self.nivel_actual], dtype=torch.float32)			


class Camion2:
	'''
	Objeto camion
	
	Atributos:
		- nombre
		- lista de bulkdrums
		- ubicación
		- ruta
		- nombres clusters
		- nombres quimicos
	'''
	def __init__(
				self, 
				nombre:str, 
				num_bulkdrums:int,
				ubicacion:int,
				ruta:Union[None, Ruta],
				nombres_clusters:List[str],
				nombres_quimicos:List[str]
			) -> None:
		self.nombre = nombre
		self.num_bulkdrums = num_bulkdrums
		self.bulkdrums = [None for _ in range(num_bulkdrums)]
		self.ubicacion = ubicacion
		self.ruta = ruta
		self.nombres_clusters = nombres_clusters
		self.num_clusters = len(nombres_clusters)
		self.nombres_quimicos = nombres_quimicos
		self.num_quimicos = len(self.nombres_quimicos)
		self.primer_bulkdrum_vacio = 0
		self.quimico_a_bulkdrum = {quimico:list() for quimico in self.nombres_quimicos}
		self.nivel_maximo = NIVEL_MAXIMO_BULKDRUM
		self.debug = False

	def __str__(self) -> str:
		cadena = '<tr>\n\t<td>Camion</td> <td>Ubicado en</td> <td>Químico</td> <td>Carga</td>\n</tr>'
		cadena += f'\n<tr>\n\t<td> {self.nombre} </td> <td>{self.nombres_clusters[self.ubicacion]}</td> '
		inicial = True
		for bulkdrum in self.bulkdrums:
			if inicial:
				cadena += str(bulkdrum) + '\n</tr> '	
				inicial = False
			else:
				cadena += ' \n<tr>\n\t<td>&nbsp;</td> <td>&nbsp;</td> ' + str(bulkdrum) + '\n</tr>'	
		return cadena

	def camion_a_tensor(self) -> torch.tensor:
		tensores = [bulkdrum.a_tensor() for bulkdrum in self.bulkdrums]
		tensor_ubicacion = torch.tensor(self.ubicacion)
		tensores += [F.one_hot(tensor_ubicacion, num_classes=self.num_clusters)]
		return torch.cat(tensores)
	
	def cargar(
				self,
				nombre_quimico:str,
				cantidad_a_cargar:float
			) -> None:
		# Determina si hay bulkdrums con el químico a bordo
		indices = self.quimico_a_bulkdrum[nombre_quimico]
		if self.debug:
			print(f'Bulkdrums donde está el químico {nombre_quimico}: {indices}')
		# Si sí, comienza a cargar y llenar los bulkdrums
		if len(indices) > 0:
			for indice in indices:
				if cantidad_a_cargar == 0:
					return None
				bulkdrum = self.bulkdrums[indice]
				if self.debug:
					print(f'\tBulkdrum {bulkdrum.nombre} tiene {bulkdrum.nivel_actual}')
				cantidad_a_cargar = bulkdrum.llenar_suavemente(cantidad_a_cargar)
				if self.debug:
					print(f'\t\tBulkdrum {bulkdrum.nombre} quedó con {bulkdrum.nivel_actual} y faltan {cantidad_a_cargar}')
		while cantidad_a_cargar > 0:
			# Aún hay que cargar químico
			if self.debug:
				print(f'Aún hay que cargar más químico {nombre_quimico}')
			if self.primer_bulkdrum_vacio < self.num_bulkdrums:
				self.quimico_a_bulkdrum[nombre_quimico].append(self.primer_bulkdrum_vacio)
				bulkdrum = BulkDrum(
					nombre=f'bulkdrum{self.primer_bulkdrum_vacio}',
					nombre_quimico=nombre_quimico,
					nivel_actual=0
				)
				cantidad_a_cargar = bulkdrum.llenar_suavemente(cantidad_a_cargar)
				self.bulkdrums[self.primer_bulkdrum_vacio] = bulkdrum
				self.primer_bulkdrum_vacio += 1
				if self.debug:
					print(bulkdrum)
			else:
				raise Exception('Error de capacidad de camion.')
		return None

	def descargar(
				self,
				nombre_quimico:str,
				cantidad_a_descargar:float
			) -> None:
		indices = self.quimico_a_bulkdrum[nombre_quimico]
		if len(indices) == 0:
			raise Exception(f'Error: Camión no lleva el químico {nombre_quimico}')
		cantidad_quimico = sum([
			bulkdrum.nivel_actual for bulkdrum in self.bulkdrums if bulkdrum.nombre_quimico == nombre_quimico
		])
		if cantidad_a_descargar > cantidad_quimico:
			raise Exception(f'Error: cantidad de {nombre_quimico} insuficiente en camión.')
		for indice in indices:
			if cantidad_a_descargar == 0:
				return None
			bulkdrum = self.bulkdrums[indice]
			cantidad_a_descargar = bulkdrum.descargar(cantidad_a_descargar)

	def a_pandas(self):
		quimicos = list()
		cantidades = list()
		for bulkdrum in self.bulkdrums:
			quimicos.append(bulkdrum.quimico.nombre)
			cantidades.append(bulkdrum.nivel_actual)
		return pd.DataFrame({'quimico':quimicos, 'cantidades':cantidades})


class Camion:
	'''
	Objeto camion
	
	Atributos:
		- nombre
		- capacidad de carga
		- lista de químicos que carga
		- ubicación
	'''
	def __init__(
				self, 
				nombre:str, 
				capacidad:float, 
				carga_quimicos:List[QuimicoEn],
				ubicacion:int,
				ruta:Union[None, Ruta],
				nombres_clusters:List[str]
			) -> None:
		self.nombre = nombre
		self.capacidad = capacidad
		self.carga_quimicos = carga_quimicos
		self.num_quimicos = len(self.carga_quimicos)
		self.ubicacion = ubicacion
		self.ruta = ruta
		self.nombres_clusters = nombres_clusters

	def __str__(self) -> str:
		cadena = '<tr>\n\t<td>Camion</td> <td>Ubicado en</td> <td>Químico</td> <td>Carga</td>\n</tr>'
		cadena += f'\n<tr>\n\t<td> {self.nombre} </td> <td>{self.nombres_clusters[self.ubicacion]}</td> '
		inicial = True
		for quimico in self.carga_quimicos:
			if inicial:
				cadena += str(quimico) + '\n</tr> '	
				inicial = False
			else:
				cadena += ' \n<tr>\n\t<td>&nbsp;</td> <td>&nbsp;</td> ' + str(quimico) + '\n</tr>'	
		return cadena

	def camion_a_tensor(
				self, 
				num_clusters:Optional[int]
			) -> torch.tensor:
		tensores = [quimico.a_tensor(en_camion=True) for quimico in self.carga_quimicos]
		tensor_ubicacion = torch.tensor(self.ubicacion)
		tensores += [F.one_hot(tensor_ubicacion, num_classes=num_clusters)]
		return torch.cat(tensores)


import typer
from pathlib import Path

from chaparral.mapa.map_objects import Mapa


IMAGEN_PAVIMENTADA = Path('data/mapas_anotados/pavimentada_prv.png')
IMAGEN_ANGOSTA = Path('data/mapas_anotados/pavimentada_angosta_prv.png')
IMAGEN_CLUSTERS = Path('data/mapas_anotados/clusters_prv.png')
IMAGEN_SIN_PAVIMENTAR = Path('data/mapas_anotados/sin_pavimentar_prv.png')
IMAGEN_BODEGA = Path('data/mapas_anotados/bodega_prv.png')

ARCHIVO_SALIDA = 'data/objetos_mapa/objetos.ndjson'


def main(nombre_archivo: str = ARCHIVO_SALIDA):
    mapa = Mapa()
    mapa.cargar_objetos_en_imagen(IMAGEN_PAVIMENTADA, 'pavimentada')
    mapa.cargar_objetos_en_imagen(IMAGEN_ANGOSTA, 'angosta')
    mapa.cargar_objetos_en_imagen(IMAGEN_CLUSTERS, 'cluster')
    mapa.cargar_objetos_en_imagen(IMAGEN_SIN_PAVIMENTAR, 'sin_pavimentar')
    mapa.cargar_objetos_en_imagen(IMAGEN_BODEGA, 'bodega')
    for i, objeto in enumerate(mapa.objetos):
        if objeto.nombre == 'VIA_PAVIMENTADA_03':
            mapa.objetos.pop(i)
    mapa.encontrar_vecinos()
    mapa.to_json(nombre_archivo)


if __name__ == "__main__":
    typer.run(main)



import pandas as pd
from pathlib import Path
import json
import numpy as np
import datetime
from typing import List, Optional, Tuple, Union

from chaparral.mapa.convenciones import POZO2CLUSTER
from chaparral.mapa.espacio_rutas import TrazadorRuta
from chaparral.modelos.constantes import (
	VALOR_CRITICO_CARGA, 
	NUM_BULKDRUMS,
	TIEMPO_DESCARGA,
)
from chaparral.modelos.modelo_utils import CampoUtils
from chaparral.modelos.clases import Camion2
from chaparral.modelos.busqueda import Nodo



class Baseline :
	"""Encuentra requerimiento de químicos y una ruta de acuerdo a heurística"""

	def __init__(self, dataframe:pd.DataFrame) -> None:
		self.dataframe = dataframe
		self.heuristicas = list()
		self.errors_log = list()
		self.max_size = 200
		self.requerimientos = None
		self.campo = None
		self.trazador = None
		self.solucion = None
		self.datos = None
		self.debug = False

	def crear_quimicos_a_camion(self) -> Tuple[List[str], pd.DataFrame]:
		"""Crea la lista de carga de químicos"""
		# Encuentra los requerimientos de químicos
		data = self.encontrar_requerimientos()
		# Inicializa el camión
		nombres_clusters = list(data.cluster.unique())
		nombres_quimicos = list(data.quimico.unique())
		camion = Camion2(
			nombre='CM1',
			num_bulkdrums=NUM_BULKDRUMS,
			ubicacion=0,
			ruta=None,
			nombres_clusters=nombres_clusters,
			nombres_quimicos=nombres_quimicos
		)
		if self.debug:
			camion.debug = True
		# Recorre los datos y va cargando los químicos en el camión
		# de acuerdo a la prioridad dias_con_stock, total_cluster, requerimiento
		clusters = ['bodega_00']
		saltar = False
		for index_, line_ in data.iterrows():
			if saltar:
				break
			cluster = line_.cluster
			if cluster in clusters:
				continue
			if self.debug:
				print(f'Tomado datos de cluster {cluster}')
			grp = data.groupby('cluster').get_group(cluster)
			for index, line in grp.iterrows():
				quimico = line.quimico
				cantidad = line.requerimiento
				if self.debug:
					print(f'Cargando {cantidad} de {quimico} para pozo {line.punto} en cluster {cluster}')
				try:
					camion.cargar(quimico, cantidad)
					clusters.append(cluster)
				except:
					# Camión lleno
					if self.debug:
						print('Camión lleno!', quimico, cluster, list(set(clusters)), len(list(set(clusters))))
					clusters = [c for c in clusters if c != cluster]
					saltar = True
					break
		# Lista de clusters a visitar
		clusters_a_visitar = list(set(clusters))
		# Guarda los requerimientos
		df = data.copy().sort_values(by=['cluster', 'punto'], inplace=False)
		df = df.drop(columns=['dias_con_stock', 'total_cluster'])
		df = pd.DataFrame(df[df['cluster'].isin(clusters_a_visitar)])
		self.requerimientos = df
		# Se toma la info del camión
		if self.debug:
			print(camion)
		df_quimicos = camion.a_pandas().sort_values(by=['quimico', 'cantidades']).reset_index().drop(columns=['index'])
		return clusters_a_visitar, df_quimicos
	
	def encontrar_ruta(
				self,
				campo,
				clusters_a_visitar:List[str]
			) -> pd.DataFrame:
		"""Encuentra la mejor ruta por los clusters"""
		# Crea el trazador de rutas
		trazador = self.crear_trazador(
			campo,
			filtro_clusters=clusters_a_visitar,
			max_size=self.max_size
		)
		self.trazador = trazador
		# Encuentra la mejor ruta
		solucion = trazador.A_star(W=1)
		# Convierte en dataframe con las columnas
		#	- hora
		#	- cluster
		if solucion is not None:
			self.solucion = solucion
			clusters_ordenados = solucion.estado + ['bodega_00']
			hora = datetime.timedelta(hours=9)
			hora_en_cluster = [hora]
			inicial = True
			for i, cluster in enumerate(clusters_ordenados[:-1]):
				siguiente_cluster = clusters_ordenados[i+1]
				if inicial:
					hora += datetime.timedelta(hours=self.campo.datos_rutas[cluster][siguiente_cluster] - TIEMPO_DESCARGA)
					inicial = False
				else:
					hora += datetime.timedelta(hours=self.campo.datos_rutas[cluster][siguiente_cluster])
				hora_en_cluster.append(hora)
			df_hora_cluster = pd.DataFrame({'Hora':hora_en_cluster, 'Cluster':clusters_ordenados})
			return df_hora_cluster
		return None

	def encuentra_hora_cluster(self, nodo:Nodo) -> Tuple[float, str]:
		return 

	def encontrar_requerimientos(self) -> pd.DataFrame:
		"""
		Retorna un dataframe de requerimientos de químicos por cluster así:
			- cluster
			- punto
			- quimico
			- requerimiento
			- días con stock
			- total cluster

		La prioridad en el orden de las filas es:
			- dias con stock
			- requerimiento total del cluster
			- requerimiento individual del punto
		"""
		# Carga datos
		data = self.dataframe.copy()
		# Deja solo los puntos activos (velocidad de consumo positiva)
		data = pd.DataFrame(data[data.velocidad_consumo > 0])
		# Prioriza por cantidad de días con stock
		data.sort_values(by='dias_con_stock', inplace=True)
		# Encuentra el cluster de cada punto
		data['cluster'] = data['punto'].apply(lambda x: self.encontrar_cluster(x))
		errores = pd.DataFrame(data[data.cluster.isna()])
		if errores.shape[0] > 0:
			self.errors_log.append(errores)
		data = pd.DataFrame(data[data.cluster.notna()])
		# Determina el requerimiento de químico de acuerdo a valor critico
		data['requerimiento'] = data['velocidad_consumo'] * VALOR_CRITICO_CARGA - data['nivel_actual']
		data['requerimiento'] = data['requerimiento'].apply(lambda x: x if x <= 250 else 250)
		# Deja solo los puntos cuyo requerimiento es positivo
		data = data[data['requerimiento'] > 0].reset_index()
		# Deja solo aquellos puntos cuya existencia sea menor al valor crítico
		data = data[data['dias_con_stock'] < VALOR_CRITICO_CARGA].reset_index()
		# Deja solo las columnas que nos interesan
		columnas_a_quitar = [columna for columna in data.columns if columna not in ['quimico','requerimiento', 'dias_con_stock', 'punto','cluster']]
		data.drop(columns=columnas_a_quitar, inplace=True)
		data = data[['cluster', 'punto', 'quimico', 'requerimiento', 'dias_con_stock']]
		# Encuentra el requerimiento total por cluster
		data['total_cluster'] = data.groupby('cluster')['dias_con_stock'].transform('mean')
		data.sort_values(
			by=[
				'total_cluster', 
				'cluster', 
				'dias_con_stock', 
				'requerimiento'
			], 
			ascending=[True, True, True, False],
			inplace=True
		)
		return data
	
	def cargar_datos(self):
		# Carga datos desde archivo
		archivo_path = self.path_archivo / Path(self.archivo)
		data = pd.read_json(archivo_path, lines=True)
		self.datos = data.copy()
		return data

	def encontrar_cluster(self, punto:str):
		try:
			return POZO2CLUSTER[punto]
		except:
			return np.nan
	
	def crear_trazador(
				self,
				campo,
				filtro_clusters:List[str],
				max_size:Optional[int]=200
			) -> TrazadorRuta:
		"""Crea el trazador de rutas"""
		cutil = CampoUtils()
		campo = cutil.cargar_campo(
			campo,
			filtro_clusters=filtro_clusters
		)
		trazador = TrazadorRuta(campo)
		trazador.max_size = max_size
		self.campo = campo
		return trazador
		
	def aplicar_heuristicas(self, df):
			"""Aplica las heurísticas sobre el dataframe"""
			for heuristica in self.heuristicas:
				df, resultado = heuristica(df)
				if resultado['responder']:
					return df
			
	def heuristica_viaje_simple(self, df):
		"""Verifica si todo cabe en un solo camion"""
		resultado = dict()
		num_quimicos = df.quimico.unique().shape
		print('->', num_quimicos)
		num_quimicos = num_quimicos[0]
		requerimientos_sobre_250 = df.requerimiento.apply(
			lambda x: x > 250
		)
		print('-->', requerimientos_sobre_250)
		requerimientos_sobre_250 = np.any(requerimientos_sobre_250)
		if num_quimicos <= 8 and not requerimientos_sobre_250:
			resultado['responder'] = True
		return df, resultado
	
	def heuristica_prioritarios(self, data):
		"""Atiende primero los pozos con menor días con stock"""
		# Ordenamos los datos por prioridad
		#	- menor cantidad de días con stock
		#	- mayor cantidad de químicos requeridos en el cluster
		#	- requerimiento individual por químico
		data.sort_values(
			by=['dias_con_stock', 'total_cluster', 'requerimiento'], 
			ascending=[True, False, False],
			inplace=True
		)
		for index, line in data.iterrows():
			cluster = line.cluster
			quimicos_cluster = data[data['cluster'] == cluster].tolist()
			print(cluster)
			print(quimicos_cluster)



import numpy as np
from collections import namedtuple
from typing import Optional

Tupla = namedtuple('Tupla', ['elemento', 'valor'])

class ListaPrioritaria:
	"""
	Estructura de datos para mantener una pila
	ordenada con base en un valor que tiene cada
	elemento.
	"""
   
	def __init__(self, max_size:Optional[int]=np.inf):
		self.pila = []
		self._index = 0
		self.max_size = max_size
				
	def __len__(self):
		return len(self.pila)

	def __iter__(self):
		return self

	def __next__(self):
		if self._index < len(self.pila):
			item = self.pila[self._index]
			self._index += 1
			return item[0]
		else:
			raise StopIteration
			
	def push(self, elemento, valor):
		tupla = Tupla(elemento, valor)
		self.pila.append(tupla)
		self.pila.sort(key=lambda x: x[1])
		if len(self.pila) > self.max_size:
			self.pila = self.pila[:self.max_size]
			
	def pop(self, with_value=False):
		if with_value:
			return self.pila.pop(0)
		else:
			return self.pila.pop(0)[0]

	def is_empty(self):
		return len(self.pila) == 0

	def __len__(self):
		return len(self.pila)

	def __str__(self):
		cadena = '['
		inicial = True
		for elemento, valor in self.pila:
			if inicial:
				cadena += '(' + str(elemento) + ',' + str(valor) + ')'
				inicial = False
			else:
				cadena += ', (' + str(elemento) + ',' + str(valor) + ')'
		return cadena + ']'
	

class Nodo:
	"""Clase para crear los nodos del árbol de búsqueda"""

	def __init__(self, estado, madre, accion, costo_camino, codigo):
		self.estado = estado
		self.madre = madre
		self.accion = accion
		self.costo_camino = costo_camino
		self.codigo = codigo

	def solucion(self):
		if self.madre == None:
			return []
		else:
			return self.solucion(self.madre) + [self.accion]

	def depth(self):
		if self.madre == None:
			return 0
		else:
			return self.depth(self.madre) + 1

	def camino_codigos(self):
		if self.madre == None:
			return [self.codigo]
		else:
			return self.camino_codigos(self.madre) + [self.codigo]

	def is_cycle(self):
		codigos = self.camino_codigos(self)
		return len(set(codigos)) != len(codigos)




'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Clase con el espacio de estados de rutas.
-----------------------------------------------
'''
import numpy as np
from collections import namedtuple
import matplotlib.pyplot as plt
import networkx as nx
import matplotlib.patches as patches
import copy
from tqdm.auto import tqdm
from typing import List, Dict, Optional, Tuple, Union

from chaparral.modelos.busqueda import ListaPrioritaria, Nodo
from chaparral.modelos.constantes import TIEMPO_DESCARGA


class RutasClusters:
	'''
	Espacio de estados para planear el camino más corto
	de un cluster a otro.
	'''
	def __init__(
				self, 
				cluster_inicial: str,
				cluster_objetivo: str,
				rutas: Dict[str,List],
				coordenadas: Dict[str, Tuple[float,float]],
				tiempo_desplazamiento: Dict[str, float]
			) -> None:
		self.cluster_inicial = cluster_inicial
		self.cluster_objetivo = cluster_objetivo
		self.rutas = rutas
		self.coordenadas = coordenadas
		self.tiempo_desplazamiento = tiempo_desplazamiento
		# Maximum size of frontera
		self.max_size = 1000	
		# Render mode
		self.render_mode = 'rgb_array'

	def _hallar_distancia(self, x, y):
		return np.sqrt((x[0] - y[0])**2 + (x[1] - y[1])**2)
        
	def render(self):
		G = nx.Graph()
		G.add_nodes_from(self.coordenadas.keys())
		list_edges = [[(cluster, x, {'distancia':self.tiempo_desplazamiento[cluster][x]}) for x in self.rutas[cluster]] for cluster in self.rutas.keys()]
		list_edges = [item for sublist in list_edges for item in sublist]
		G.add_edges_from(list_edges, color='red')
		order = {cluster:self.coordenadas[cluster] for cluster in self.coordenadas.keys()}
		options = {
			'node_color': 'green',
			'node_size': 10,
			'verticalalignment':'bottom',
			'edge_color': 'red',
			'width': 0.25,
			'font_size': 10
		}
		fig = plt.figure(figsize=(15,15))
		ax = fig.add_subplot()
		nx.draw_networkx(G, pos=order, font_weight='bold', **options)
		edge_labels = nx.get_edge_attributes(G, 'distancia')
		nx.draw_networkx_edge_labels(G, order, edge_labels)
		ax.axis('off')
		fig.canvas.draw()
		string = fig.canvas.renderer.buffer_rgba()
		array = np.array(string)
		if self.render_mode == 'rgb_array':
			return array
		else:
			plt.show()

	def acciones_aplicables(self, estado):
		# Devuelve una lista de ciudades a las que se puede llegar
		# desde la ciudad descrita en estado
		# Input: estado, nombre de una ciudad
		# Output: lista de ciudades
		return self.rutas[estado]

	def transicion(self, estado, accion):
		# Devuelve la ciudad a la que se va
		# Input: estado, que es una ciudad
		#        accion, que es una ciudad
		# Output: accion, que es una ciudad
		return accion

	def test_objetivo(self, estado):
		# Devuelve True/False dependiendo si el estado
		# resuelve el problema
		# Input: estado, que es una ciudad
		# Output: True/False
		return estado == self.cluster_objetivo

	def costo(self, estado, accion):
		# Devuelve la distancia desde estado hasta accion (que es una ciudad)
		# Input: 
		#        estado, que es una ciudad
		#        accion, que es una ciudad
		# Output: distancia de acuerdo al diccionario rutas más tiempo de descarga
		return self.tiempo_desplazamiento[estado][accion] + TIEMPO_DESCARGA

	def codigo(self, estado):
		return estado

	def nodo_hijo(self, madre, accion):
		# Método para crear un nuevo nodo
		# Input: problema, que es un objeto de clase ocho_reinas
		#        madre, que es un nodo,
		#        accion, que es una acción que da lugar al estado del nuevo nodo
		# Output: nodo
		estado = self.transicion(madre.estado, accion)
		costo_camino = madre.costo_camino + self.costo(madre.estado, accion)
		codigo = self.codigo(estado)
		return Nodo(estado, madre, accion, costo_camino, codigo)

	def dijkstra(self):
		s = self.cluster_inicial
		cod = self.codigo(s)
		nodo = Nodo(s, None, None, 0, cod)
		frontera = ListaPrioritaria()
		frontera.push(nodo, 0)
		explorados = {}
		explorados[cod] = 0
		while not frontera.is_empty():
			nodo = frontera.pop()
			if self.test_objetivo(nodo):
				return nodo
			for a in self.acciones_aplicables(nodo.estado):
				hijo = self.nodo_hijo(nodo, a)
				s = hijo.estado
				cod = hijo.codigo
				c = hijo.costo_camino
				if (cod not in explorados.keys()) or (c < explorados[cod]):
					frontera.push(hijo, c)
					assert(len(frontera) < self.max_size), 'MAX_SIZE superado'
					explorados[cod] = c
		return None

	def A_star(
				self, 
				W:float, 
				verbose:bool=False
			) -> Nodo:
		s = self.cluster_inicial
		v = self.heuristica(s)
		cod = self.codigo(s)
		nodo = Nodo(s, None, None, 0, cod)
		frontera = ListaPrioritaria()
		frontera.push(nodo, v)
		explorados = {}
		explorados[cod] = W * v
		while not frontera.is_empty():
			if verbose:
				print(f'Tamaño frontera:{len(frontera)} --- min:{frontera.pila[0][1]}')
			nodo = frontera.pop()
			if self.test_objetivo(nodo.estado):
				return nodo
			for a in self.acciones_aplicables(nodo.estado):
				hijo = self.nodo_hijo(nodo, a)
				s = hijo.estado
				v = self.heuristica(s)
				cod = hijo.codigo
				c = hijo.costo_camino
				h = W * v + c
				if (cod not in explorados.keys()) or (h < explorados[cod]):
					frontera.push(hijo, h)
					explorados[cod] = h
		return None
	
	def heuristica(self, cluster:str) -> float:
		return self._hallar_distancia(
			self.coordenadas[self.cluster_objetivo], 
			self.coordenadas[cluster]
		)



class TrazadorRuta:
    """Trazador de rutas"""
    def __init__(self, campo):
        self.cluster_inicial = ['bodega_00']
        self.rutas = campo.datos_rutas
        self.coords = campo.coordenadas
        self.clusters = sorted([cluster.nombre for cluster in campo.clusters])
        self.G = None
        self.max_size = 1000

    def render(self, estado):
        """ Creacion y plot del grafo como vertices las localidades y pesos el valor de la distancia entre cada para de localidades """
        self.G = nx.Graph()
        n = len(estado)
        for i in range(n):
            x, y = self.coords[estado[i]]
            self.G.add_node(estado[i], pos = (x,y))
        for i in range(n-1):
            self.G.add_edge(estado[i], estado[i+1], weight = round(self.rutas[estado[i]][estado[i+1]], 2))
        pos = nx.get_node_attributes(self.G, 'pos')
        pesos = nx.get_edge_attributes(self.G,'weight')
        plt.figure(figsize=(12,12))
        nx.draw_networkx(self.G, pos)
        nx.draw_networkx_edge_labels(self.G, pos, edge_labels = pesos)
        plt.show()

    def pintar_camino(self, camino):
        # Input: lista con el camino de los estados
        # Output: plot con la forma de resolver las torres
        for estado in camino:
            clear_output(wait=True)
            self.pintar_estado(estado)
            plt.show()
            sleep(.5)

    def acciones_aplicables(self, estado):
        """ Retorna las posibles acciones dado el estado
            Input: estado (localidad)
            Output: Lista con las localidades no visitadas
        """
        return [x for x in self.clusters if x not in estado]

    def transicion(self, estado, accion):
        """ Retorna la lista actualizada con la accion realizada
            Input: estado (lista con el camino en el momento)
                   accion (desplazamiento)
            Output: Copia de la lista estado actualizada """
        lista = copy.deepcopy(estado)
        lista.append(accion)
        return lista

    def test_objetivo(self, estado):
        """ Verifica si ya fueron visitadas todas las localidades
            Input: estado (lista con el camino)
            Output: Boolean """
        return set(self.clusters) == set(estado)

    def codigo(self, estado):
        """ Actualiza el codigo
            Input: estado (lista con el camino)
            Output: cadena """
        cad = ""
        for i in estado:
            cad = cad + " - " + i
        return cad

    def costo(self, estado1, estado2):
        """ Peso entre el estado y la accion
            Input: estado (camino actual)
                   accion (desplazamiento)
            Output: Int """
        loc = self.rutas[estado1[-1]]
        try:
            costo_ = loc[estado2]
            return costo_
        except Exception as e:
            print(f'Error al buscar costo ({e})')
            print(f'{estado1[-1]} --> {estado2}')
            raise Exception()

    def dijkstra(self):
        s = self.cluster_inicial
        cod = self.codigo(s)
        nodo = Nodo(s, None, None, 0, cod)
        frontera = ListaPrioritaria(max_size=self.max_size)
        frontera.push(nodo, 0)
        explorados = {}
        explorados[cod] = 0
        pbar = tqdm(total=len(self.clusters))
        contador_estados_recorridos = 1
        while not frontera.is_empty():
            nodo = frontera.pop()
            diferencia = len(set(nodo.estado)) - contador_estados_recorridos
            if  diferencia > 0:
                pbar.update(diferencia)
                contador_estados_recorridos += diferencia 
            elif diferencia < 0:
                pbar.close()
                pbar = tqdm(total=len(self.clusters))
                pbar.update(len(set(nodo.estado)))
                contador_estados_recorridos = len(set(nodo.estado))
            if self.test_objetivo(nodo.estado):
                pbar.close()
                return nodo
            for a in self.acciones_aplicables(nodo.estado):
                hijo = self.nodo_hijo(nodo, a)
                s = hijo.estado
                cod = hijo.codigo
                c = hijo.costo_camino
                if (cod not in explorados.keys()) or (c < explorados[cod]):
                    frontera.push(hijo, c)
                    assert(len(frontera) < self.max_size + 1), 'MAX_SIZE superado'
                    explorados[cod] = c
        pbar.close()
        return None

    def A_star(
                self, 
                W:float, 
                verbose:bool=False
            ) -> Nodo:
        s = self.cluster_inicial
        v = self.heuristica(s)
        cod = self.codigo(s)
        nodo = Nodo(s, None, None, 0, cod)
        frontera = ListaPrioritaria(max_size=self.max_size)
        frontera.push(nodo, v)
        explorados = {}
        explorados[cod] = W * v
        pbar = tqdm(
            total=len(self.clusters), 
            desc='%clusters visitados',
            leave=False
        )
        contador_estados_recorridos = 1
        pbar.update(contador_estados_recorridos)
        while not frontera.is_empty():
            nodo = frontera.pop()
            diferencia = len(set(nodo.estado)) - contador_estados_recorridos
            if  diferencia > 0:
                pbar.update(diferencia)
                contador_estados_recorridos += diferencia 
            if self.test_objetivo(nodo.estado):
                return nodo
            for a in self.acciones_aplicables(nodo.estado):
                hijo = self.nodo_hijo(nodo, a)
                s = hijo.estado
                v = self.heuristica(s)
                cod = hijo.codigo
                c = hijo.costo_camino
                h = W * v + c
                if (cod not in explorados.keys()) or (h < explorados[cod]):
                    frontera.push(hijo, h)
                    explorados[cod] = h
        return None
	
    def heuristica(self, clusters:str) -> float:
        if len(clusters) < 2:
            return 0
        objetivo = clusters[-1]
        anterior = clusters[-2]
        return self._hallar_distancia(
			self.coords[objetivo], 
			self.coords[anterior]
        )

    def _hallar_distancia(self, x, y):
        return np.sqrt((x[0] - y[0])**2 + (x[1] - y[1])**2)

    def nodo_hijo(self, madre, accion):
        # Método para crear un nuevo nodo
        # Input: problema, que es un objeto de clase ocho_reinas
        #        madre, que es un nodo,
        #        accion, que es una acción que da lugar al estado del nuevo nodo
        # Output: nodo
        estado = self.transicion(madre.estado, accion)
        costo_camino = madre.costo_camino + self.costo(madre.estado, accion)
        codigo = self.codigo(estado)
        return Nodo(estado, madre, accion, costo_camino, codigo)